"""Pruner."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from intel_extension_for_transformers.optimization.utils.utility import LazyImport

torch = LazyImport('torch')
from .patterns import get_pattern
from .schedulers import get_scheduler
from .criteria import get_criterion, CRITERIA
from .regs import get_reg
from .logger import logger

PRUNERS = {}

def register_pruner(name):
    """Class decorator to register a Pruner subclass to the registry.

    Decorator function used before a Pattern subclass.
    Make sure that the Pruner class decorated by this function can be registered in PRUNERS.

    Args:
        cls (class): The subclass of register.
        name: A string. Define the pruner type.

    Returns:
        cls: The class of register.
    """

    def register(pruner):
        PRUNERS[name] = pruner
        return pruner

    return register

def parse_valid_pruner_types():
    """Get all valid pruner names."""
    valid_pruner_types = []
    for x in CRITERIA.keys():
        for p in ["", "_progressive"]:
            valid_pruner_types.append(x + p)
    valid_pruner_types.append("pattern_lock")
    return valid_pruner_types

def get_pruner(config, modules):
    """Get registered pruner class.

    Get a Pruner object from PRUNERS.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.

    Returns:
        A Pruner object.

    Raises: AssertionError: Cuurently only support pruners that have been registered in PRUNERS.
    """
    ## do the ugly work here
    if "progressive" not in config["pruning_type"]:
        name = config["pruning_type"]
        config["progressive"] = False
    else:
        # if progressive, delete "progressive" words and reset config["progressive"]
        name = config["pruning_type"][0:-12]
        config["progressive"] = True
    if name in CRITERIA:
        if config["progressive"] == False:
            config['criterion_type'] = name
            name = "basic"  ##return the basic pruner
        else:
            config['criterion_type'] = name
            name = "progressive"  ## return the progressive pruner

    if name not in PRUNERS.keys():  # pragma: no cover
        assert False, f"does not support {name}, currently only support {parse_valid_pruner_types()}"
    return PRUNERS[name](config, modules)


class BasePruner:
    """Pruning Pruner.

    The class which executes pruning process.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.

    Attributes:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.
        masks: A dict {"module_name": Tensor} that stores the masks for modules' weights.
        scores: A dict {"module_name": Tensor} that stores the score for modules' weights,
            which are used to determine what parts to be pruned by a criterion.
        pattern: A Pattern object defined in ./patterns.py
        scheduler: A scheduler object defined in ./scheduler.py
        current_sparsity_ratio: A float representing the current model's sparsity ratio; it is initialized to be zero.
        global_step: An integer representing the total steps the model has run.
        start_step: An integer representing when to trigger pruning process.
        end_step: An integer representing when to end pruning process.
        pruning_frequency: An integer representing the pruning frequency; it is valid when iterative
            pruning is enabled.
        target_sparsity_ratio: A float showing the final sparsity after pruning.
        max_sparsity_ratio_per_op: A float showing the maximum sparsity ratio for every module.
    """

    def __init__(self, config, modules):
        """Initialize."""
        self.modules = modules
        self.config = config
        self.masks = {}
        self.global_step = 0
        self.handled_global_step = -1
        self.start_step = self.config['start_step']
        self.end_step = self.config['end_step']
        self.pruning_frequency = self.config['pruning_frequency']
        ##this is different with original code
        self.total_prune_cnt = (self.end_step - self.start_step + self.pruning_frequency) \
                               // self.pruning_frequency
        self.completed_pruned_cnt = 0
        self.total_prune_cnt -= 1  ## not pruning at step 0
        if self.total_prune_cnt == 0:  # pragma: no cover
            self.total_prune_cnt = 1
            self.completed_pruned_cnt = 1

        for key in self.modules.keys():
            module = self.modules[key]
            self.masks[key] = torch.ones(module.weight.shape).to(module.weight.device)  ##TODO support bias or others

        self.target_sparsity_ratio = self.config['target_sparsity']
        self.current_sparsity_ratio = 0.0
        self.init_sparsity_ratio = 0.0
        self._init()

    def _init(self):  # pragma: no cover
        """Auxiliary function for initializing."""
        pass

    def on_epoch_begin(self, epoch):
        """Implement at the beginning of each epoch."""
        pass

    def mask_weights(self):
        """Apply masks to corresponding modules' weights.
        
        Weights are multipled with masks. This is the formal pruning process.
        """
        with torch.no_grad():
            for key in self.modules.keys():
                module = self.modules[key]
                module.weight.data = module.weight.data * self.masks[key]

    def mask_weights_general(self, input_masks):
        """Apply input masks to corresponding modules' weights.
        
        Weights are multipled with input_masks. 

        Args:
            input_masks: A dict {"module_name": Tensor} that stores the masks for modules' weights.
        """
        with torch.no_grad():
            for key in self.modules.keys():
                module = self.modules[key]
                module.weight.data = module.weight.data * input_masks[key]

    def on_step_begin(self, local_step):
        """Implement at the start of each step."""
        if self.handled_global_step == self.global_step:  # pragma: no cover
            return
        self.update_masks(local_step)
        self.handled_global_step = self.global_step

    def update_masks(self, local_step):  # pragma: no cover
        """Update the masks at a given local step."""
        pass

    def on_epoch_end(self):  # pragma: no cover
        """Implement at the end of each epoch."""
        pass

    def on_step_end(self):  # pragma: no cover
        """Implement at the end of each step."""
        pass

    def on_before_optimizer_step(self):  # pragma: no cover
        """Implement before optimizer.step()."""
        pass

    def on_after_optimizer_step(self):  # pragma: no cover
        """Implement after optimizer.step().
        
        Prune the model after optimization.
        """
        self.mask_weights()
        self.global_step += 1

    def on_train_begin(self):  # pragma: no cover
        """Implement at the beginning of training phase."""
        pass

    def on_train_end(self):  # pragma: no cover
        """Implement at the end of training phase."""
        pass
 
    def on_before_eval(self):  # pragma: no cover
        """Implement at the beginning of evaluation phase."""
        pass

    def on_after_eval(self):  # pragma: no cover
        """Implement at the end of evaluation phase."""
        pass

    def check_is_pruned_step(self, step):  
        """Check if a pruning process should be performed at the current step.
        
        Args:
            step: an integer representing the number of current step.
            
        Returns: 
            A Boolean.
        """
        if step < self.start_step or step > self.end_step:
            return False
        if int(step - self.start_step) % self.pruning_frequency == 0:
            return True
        return False


@register_pruner("basic")
class BasicPruner(BasePruner):
    """Pruning Pruner.

    The class which executes pruning process.
    1. Defines pruning functions called at step begin/end, epoch begin/end.
    2. Defines the pruning criterion.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.

    Attributes:
        pattern: A Pattern object that defines pruning weights' arrangements within space.
        criterion: A Criterion Object that defines which weights are to be pruned
        scheduler: A Scheduler object that defines how the model's sparsity changes as training/pruning proceeds.
        reg: A Reg object that defines regulization terms.
    """

    def __init__(self, config, modules):
        """Initialize."""
        # self.modules = modules
        # self.config = config
        # self.masks = {}
        super(BasicPruner, self).__init__(config, modules)

    def _init(self):
        """Auxiliary function for initializing."""
        self.pattern = get_pattern(self.config, self.modules)
        self.scheduler = get_scheduler(self.config)
        self.criterion = get_criterion(self.config, self.modules)
        self.reg = get_reg(self.config, self.modules, self.pattern)
        # if switch off progressive but use per-channel pruning, give a warn
        if "channel" in self.pattern.pattern:
            logger.info("UserWarning: use per-channel pruning pattern without progressive pruning!")
            logger.info("Instead, enabling progressive pruning would be a better choice.")
        else:
            pass

    def set_global_step(self, global_step):  # pragma: no cover
        """Set global step number."""
        self.global_step = global_step

    # def on_step_begin(self, local_step):
    #     """Implement at the start of each step.
    #
    #     Update the masks at a given local_step.
    #     """
    #     self.update_masks(local_step)

    def update_masks(self, local_step):  # pragma: no cover
        """Update the masks at a given local step."""
        if self.global_step == self.start_step:
            if self.config['lock_init_sparsity']:
                self.masks = self.pattern.get_pattern_lock_masks(self.modules)
                self.init_sparsity_ratio = self.pattern.get_sparsity_ratio(self.masks)
                self.current_sparsity_ratio = self.init_sparsity_ratio

        if not self.check_is_pruned_step(self.global_step):
            return

        if self.current_sparsity_ratio > self.target_sparsity_ratio:
            return

        self.criterion.on_step_begin()
        current_target_sparsity_ratio = self.scheduler.update_sparsity_ratio(self.target_sparsity_ratio,
                                                                             self.completed_pruned_cnt,
                                                                             self.total_prune_cnt, self.masks,
                                                                             self.init_sparsity_ratio)
        logger.info(f"current target ratio is {current_target_sparsity_ratio}")

        self.completed_pruned_cnt += 1
        if self.criterion.scores == {}:
            return
        self.masks = self.pattern.get_masks(self.criterion.scores, current_target_sparsity_ratio, self.masks)
        self.mask_weights()

        self.current_sparsity_ratio = self.pattern.get_sparsity_ratio(self.masks)
        logger.info(f"current sparsity ratio is {self.current_sparsity_ratio}")

    def on_before_optimizer_step(self):
        """Implement before optimizer.step()."""
        self.reg.on_before_optimizer_step()
        self.criterion.on_before_optimizer_step()

    def on_after_optimizer_step(self):
        """Prune the model after optimization."""
        ##the order of the following three lines can't not be exchanged
        if self.global_step >= self.start_step and self.global_step <= self.end_step:
            self.reg.on_after_optimizer_step()
        self.mask_weights()

        self.global_step += 1


@register_pruner('pattern_lock')
class PatternLockPruner(BasePruner):  # pragma: no cover
    """Pruning Pruner.

    A Pruner class derived from BasePruner.
    In this pruner, original model's sparsity pattern will be fixed while training.
    This pruner is useful when a user trains a sparse model without changing its original structure.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.

    Attributes:
        Inherit from parent class Pruner.
    """

    def __init__(self, config, modules):
        """Initialize."""
        super(PatternLockPruner, self).__init__(config, modules)
        self.pattern = get_pattern(self.config, modules)
        assert self.config.end_step == self.config.start_step, "pattern_lock pruner only supports one shot mode"

    def update_masks(self, local_step):
        """Update the masks at a given local step."""
        if not self.check_is_pruned_step(self.global_step):
            return
        self.masks = self.pattern.get_pattern_lock_masks(self.modules)

    def on_after_optimizer_step(self):
        """Implement after optimizer.step().
        
        Prune the model after optimization.
        """
        self.mask_weights()
        self.global_step += 1


@register_pruner('progressive')
class ProgressivePruner(BasicPruner):
    """Pruning Pruner.

    A Pruner class derived from BasePruner. In this pruner, mask interpolation will be applied.
    Mask interpolation is a fine-grained improvement for NxM structured pruning by adding interval 
        masks between masks of two pruning steps.

    Args:
        modules: A dict {"module_name": Tensor} that stores the pruning modules' weights.
        config: A config dict object that contains the pruner information.

    Attributes:
        Inherit from parent class Pruner.
    """

    def __init__(self, config, modules):
        """Initialize."""
        super(ProgressivePruner, self).__init__(config, modules)

    def _init(self):
        """Auxiliary function for initialization."""
        self.pattern = get_pattern(self.config, self.modules)
        self.scheduler = get_scheduler(self.config)
        self.criterion = get_criterion(self.config, self.modules)
        self.reg = get_reg(self.config, self.modules, self.pattern)
        # progressive pruning set up, including check up paramters.
        self.use_progressive = self.config["progressive"]
        # progressive parameters
        # dict passed to Pattern's functions
        self.progressive_configs = {
            "progressive_steps": 4,
            "progressive_type": "scores",
            "use_global": True
        }
        self.progressive_steps = self.progressive_configs["progressive_steps"]
        self.progressive_type = self.progressive_configs["progressive_type"]
        self.use_global = self.progressive_configs["use_global"]
        self.progressive_logger = False
        self._init_for_progressive()

    def _init_for_progressive(self):
        """Auxiliary function for initializing progressive pruning."""
        # detailed progressive parameters will stored at patterns.py
        # step 1: check if pattern is NxM
        if "x" not in self.pattern.pattern:
            raise NotImplementedError(f"Currently progressive only " \
                                      f"support NxM and per-channel pruning patterns.")

        # step 2: check if current set up will "degrade" into non-progressive
        degrading_flag = False
        if (self.end_step - self.start_step) <= self.progressive_steps or self.progressive_steps <= 1:
            logger.info("Current progressive setting will degrading to non-progressive pruning.")
            self.use_progressive = False
            return

        # step 3: log hyper-parameters. and check validity.
        if self.use_progressive:
            logger.info(f"Progressive pruning is enabled!")
            logger.info(f"Progressive pruning steps: {self.progressive_steps}")
            logger.info(f"Progressive type: {self.progressive_type}")
            logger.info(f"Progressive balance: {self.use_global}")
            self.check_progressive_validity()
            self.pre_masks = copy.deepcopy(self.masks)
            self.progressive_masks = copy.deepcopy(self.masks)
            if self.pruning_frequency < self.progressive_steps:  ##TODO trick
                self.progressive_steps = self.pruning_frequency
                # if self.progressive_steps == 3:
                #     self.progressive_steps = 2
                self.pruning_frequency_progressive = self.progressive_steps
            else:
                self.pruning_frequency_progressive = self.pruning_frequency // self.progressive_steps
            # this is a structural pruning step, it fits self.pruning_frequency
            self.structured_update_step = 0

    def check_progressive_validity(self):  # pragma: no cover
        """Check if the settings of progressive pruning are valid."""
        # check some problematic settings
        if self.progressive_type == "linear":
            if self.use_global:
                # when global progressive is applied, linear type is contradict.
                raise NotImplementedError("Global progressive pruning do not support linear pattern")
            # When linear, progressive_step should not meet a indivisible 
            for key in self.pattern.block_size.keys():
                block_size = self.pattern.block_size[key]
                progressive_direction = max(block_size)
                if progressive_direction % self.progressive_steps != 0:
                    raise ValueError(
                        f"In layer {key}, its pruning pattern is {block_size}, " \
                        f"while progressive steps {self.progressive_steps} is indivisible.")
        else:
            for key in self.pattern.block_size.keys():
                block_size = self.pattern.block_size[key]
                total_block_size = block_size[0] * block_size[1]
                if total_block_size < self.progressive_steps:
                    raise ValueError(
                        f"In layer {key}, its pruning pattern is {block_size}, " \
                        f"while progressive steps {self.progressive_steps} is overflowing.")

    def check_is_pruned_progressive_step(self, step):
        """Check if a progressive pruning process should be performed at the current step.
        
        Args:
            step: an integer representing the number of current step.
            
        Returns: 
            A Boolean.
        """
        # used in progressive pruning
        if step < self.start_step or step > self.end_step:
            return False
        if int(step - self.start_step) % self.pruning_frequency_progressive == 0:
            return True
        return False

    def update_masks_progressive(self, local_step):
        """Update the masks in progressive pruning mode at a given local step."""
        if self.global_step == self.start_step:
            if self.config['lock_init_sparsity']:  # pragma: no cover
                self.masks = self.pattern.get_pattern_lock_masks(self.modules)
                self.init_sparsity_ratio = self.pattern.get_sparsity_ratio(self.masks)
                self.current_sparsity_ratio = self.init_sparsity_ratio

        # case 1: step is not in [start_step, end_step] or it is not either pruning or progressive pruning step.
        if (self.check_is_pruned_step(self.global_step) == False) and (
                self.check_is_pruned_progressive_step(self.global_step) == False):
            return
        if self.current_sparsity_ratio > self.target_sparsity_ratio:
            return

        # case 2: step which does progressive update, but it is not a pruning step in case 3
        if self.check_is_pruned_progressive_step(self.global_step) \
                and self.check_is_pruned_step(self.global_step) == False:
            # do not do global pruning, only do the progressive mask update.
            step_offset = self.global_step - self.structured_update_step
            progressive_idx = step_offset // self.pruning_frequency_progressive
            if progressive_idx < (self.progressive_steps - 1):
                self.progressive_masks = self.pattern.update_progressive_masks(self.pre_masks, self.masks, \
                                                                               self.criterion.scores, \
                                                                               progressive_idx + 1, \
                                                                               self.progressive_configs)
            else:
                # in the end, directly use new masks.
                for n in self.masks.keys():
                    self.progressive_masks[n] = self.masks[n].clone()
            self.mask_weights_general(self.progressive_masks)
            if self.progressive_logger:
                self.print_progressive_sparsity()
            return

        # case 3: a pruning step, generate new masks, progressive masks also update.
        tmp_step = self.global_step
        self.structured_update_step = tmp_step
        current_target_sparsity_ratio = self.scheduler.update_sparsity_ratio(self.target_sparsity_ratio,
                                                                             self.completed_pruned_cnt,
                                                                             self.total_prune_cnt, self.masks)
        logger.info(f"current target ratio is {current_target_sparsity_ratio}")
        self.criterion.on_step_begin()
        self.completed_pruned_cnt += 1
        if self.criterion.scores == {}:
            return
        for n in self.masks.keys():
            self.pre_masks[n] = self.masks[n].clone()
        # update new masks
        self.masks = self.pattern.get_masks(self.criterion.scores, current_target_sparsity_ratio, self.masks, )
        self.progressive_masks = self.pattern.update_progressive_masks(self.pre_masks, self.masks, \
                                                                       self.criterion.scores, 1, \
                                                                       self.progressive_configs)
        self.mask_weights_general(self.progressive_masks)
        if self.progressive_logger:
            self.print_progressive_sparsity()
        return

    def on_step_begin(self, local_step):
        """Update the masks at a given local_step.
        
        Implement at the start of each step.
        """
        if self.handled_global_step == self.global_step:
            return

        if not self.use_progressive:  # pragma: no cover
            # As _init_for_progressive() works, when degrades to non-progressive
            # just call BasicPruner's update_masks().
            self.update_masks(local_step)
        else:
            self.update_masks_progressive(local_step)
        self.handled_global_step = self.global_step

    def on_before_optimizer_step(self):
        """Implement before optimizer.step()."""
        self.reg.on_before_optimizer_step()
        self.criterion.on_before_optimizer_step()

    def on_after_optimizer_step(self):
        """Prune the model after optimization."""
        ##the order of the following three lines can't not be exchanged
        if self.global_step >= self.start_step and self.global_step <= self.end_step:
            self.reg.on_after_optimizer_step()
        if not self.use_progressive:  # pragma: no cover
            self.mask_weights()
        else:
            self.mask_weights_general(self.progressive_masks)

        self.global_step += 1

    def print_progressive_sparsity(self):  # pragma: no cover
        """Output the progressive sparsity."""
        cur_sp = self.pattern.get_sparsity_ratio_progressive(self.progressive_masks)
        logger.info("Step: {} -> Current progressive sparsity: {}".format(self.global_step, cur_sp))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The EinsumwithArange Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
import copy


@pattern_registry(pattern_type='EinsumwithArange')
class EinsumwithArange(Pattern):
    """The InnerproductReshapeFusion pattern.

    Fuse the original sub-graph into the custom acceleration 'EinsumwithArange' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'EinsumwithArange': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Arange'), (2, 'Einsum')]
                                ],
                        'out': [[(0, 'Range'), (1, 'Reshape'), (2, 'Matmul')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 0,
                        2: 2
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        }
                        ], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            2: [0]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            2: [0]
                        }], [[0], 1]],
                    },
                    'returns': [0, 1, 2]
                },
                
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (2, 'Add'), (3, 'Arange'), (4, 'Einsum')],
                               [(), (1, 'Shape'), (2, 'Add')]
                                ],
                        'out': [[(0, 'Range'), (1, 'Reshape'), (2, 'Matmul')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                        1: 0,
                        2: 4
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        },{
                            'input_data': [2]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]],
                        2: [[{
                            4: [1]
                        }], [[1], 2]]
                        
                        
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            4: [0]
                        }], [[0], 1]],
                    },
                    'returns': [0]
                }
                
            ]
        }
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                range_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                if 'end' in ret_old_nodes[i][0].attr.keys():
                    attr['end_with_shape'] = ret_old_nodes[i][0].attr['end']
                model.nodes[range_node_idx].attr = attr
                
                matmul_node = model.get_node_by_name(new_node_names[i][2])
                reshape_node = model.get_node_by_name(new_node_names[i][1])
                reshape_node.attr = OrderedDict({'dst_shape': '-1, 1'})
                reshape_output = Tensor(name=reshape_node.input_tensors[0].name + "_reshape",
                                        source_op=[matmul_node.name + "_reshape"],
                                        dest_op=[matmul_node.name],
                                        dtype=matmul_node.input_tensors[0].dtype)
                matmul_node.input_tensors[1].dest_op = [matmul_node.name + "_reshape"]
                range_2 = model.get_node_by_name(matmul_node.input_tensors[1].source_op[0])
                range_2.output_tensors[0].dest_op = [matmul_node.name + "_reshape"]
                reshape_op = util.construct_node(
                    node_name=matmul_node.name + "_reshape",
                    op_type='Reshape',
                    input_tensors=[matmul_node.input_tensors[1]],
                    output_tensors=[reshape_output],
                    attr=OrderedDict({'dst_shape': '1, -1'}))
                
                matmul_node.input_tensors[1] = reshape_output
                insert_idx = model.get_node_id(new_node_names[i][2])
                model.insert_nodes(insert_idx, [reshape_op])
                
        pattern_dict = pattern_mapping_config['EinsumwithArange'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("EinsumwithArange", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr(new_node_names, ret_old_nodes, model)

        def _set_attr1(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                range_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                attr['algorithm'] = "add"
                attr['end_with_shape'] = 1
                model.nodes[range_node_idx].attr = attr            
                
                matmul_node = model.get_node_by_name(new_node_names[i][2])
                reshape_node = model.get_node_by_name(new_node_names[i][1])
                reshape_node.attr = OrderedDict({'dst_shape': '-1, 1'})
                reshape_output = Tensor(name=reshape_node.input_tensors[0].name + "_reshape",
                                        source_op=[matmul_node.name + "_reshape"],
                                        dest_op=[matmul_node.name],
                                        dtype=matmul_node.input_tensors[0].dtype)
                matmul_node.input_tensors[1].dest_op = [matmul_node.name + "_reshape"]
                range_2 = model.get_node_by_name(matmul_node.input_tensors[1].source_op[0])
                range_2.output_tensors[0].dest_op = [matmul_node.name + "_reshape"]
                reshape_op = util.construct_node(
                    node_name=matmul_node.name + "_reshape",
                    op_type='Reshape',
                    input_tensors=[matmul_node.input_tensors[1]],
                    output_tensors=[reshape_output],
                    attr=OrderedDict({'dst_shape': '1, -1'}))
                matmul_node.input_tensors[1] = reshape_output
                insert_idx = model.get_node_id(new_node_names[i][2])
                model.insert_nodes(insert_idx, [reshape_op])
                
        pattern_dict = pattern_mapping_config['EinsumwithArange'][1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("EinsumwithArange", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr1(new_node_names, ret_old_nodes, model)


            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The QuantizedGraphDtypeRefactor Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from ..ops import Tensor
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
import copy


@pattern_registry(pattern_type='QuantizedGraphDtypeRefactor')
class QuantizedGraphDtypeCheck(Pattern):
    """The QuantizedGraphDtypeRefactor pattern.

    Fuse the original sub-graph into the custom acceleration 'QuantizedGraphDtypeRefactor' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""

        def _quantized_dtype(model):
            dtype = 'fp32'
            for node in model.nodes:
                if node.op_type == 'Quantize':
                    dtype = node.attr['output_dtype']
                    break
            return dtype

        def _get_dst_ops(node, model):
            ret = []
            output_name = node.output_tensors[0].name
            for node in model.nodes:
                for input_tensor in node.input_tensors:
                    if output_name == input_tensor.name:
                        ret.append(node)
            return ret

        def _insert_quant_node(pre_node, model, graph_dtype):
            output_tensor = copy.deepcopy(pre_node.output_tensors[0])
            output_tensor.dtype = graph_dtype
            pre_node.output_tensors[0].name = pre_node.output_tensors[0].name + '_before_quant'
            pre_node.output_tensors[0].dest_op = [pre_node.name + '_quant']
            input_tensor = copy.deepcopy(pre_node.output_tensors[0])
            quantize_op = util.construct_node(node_name=pre_node.name + "_quant",
                                              op_type='Quantize',
                                              input_tensors=[input_tensor],
                                              output_tensors=[output_tensor],
                                              attr=OrderedDict({'output_dtype': graph_dtype}))
            insert_idx = model.get_node_id(pre_node.name)
            model.insert_nodes(insert_idx + 1, [quantize_op])
            return model

        def _check_dst_op(start_node, model, checker, graph_dtype):
            dst_ops = _get_dst_ops(start_node, model)
            if len(dst_ops) == 0:
                return
            for op in dst_ops:
                if EXECUTOR_TYPE.get(op.op_type, op.op_type) in checker[graph_dtype]:
                    if op.attr.get('output_dtype', None) != graph_dtype:
                        op.attr['output_dtype'] = graph_dtype
                        _check_dst_op(op, model, checker, graph_dtype)
                else:
                    continue

        def _scan_nodes_graph_dtype(model, checker, graph_dtype):
            for node in model.nodes:
                if node.attr and node.attr.get('output_dtype', 'fp32') == graph_dtype:
                    dst_ops = _get_dst_ops(node, model)
                    for op in dst_ops:
                        if EXECUTOR_TYPE.get(op.op_type, op.op_type) in checker[graph_dtype] and \
                            op.attr.get('output_dtype', 'fp32') != graph_dtype:
                            op.attr['output_dtype'] = graph_dtype

        def _remove_redundant_quant_node(model, graph_dtype):
            remove_node_names = []
            for node in model.nodes:
                if node.op_type == 'Quantize':
                    pre_node = None
                    try:
                        pre_node = model.get_node_by_name(node.input_tensors[0].source_op[0])
                    except:
                        pre_node = None
                    if pre_node and pre_node.attr.get('output_dtype', None) == graph_dtype:
                        dst_ops = _get_dst_ops(pre_node, model)
                        if len(dst_ops) == 1 and dst_ops[0].op_type == "Quantize":
                            pre_node.output_tensors = copy.deepcopy(node.output_tensors)
                            remove_node_names.append(node.name)
                        else:
                            if len(dst_ops) == 2 and 'Output' in [o.op_type for o in dst_ops]:
                                t_name = pre_node.output_tensors[0].name
                                pre_node.output_tensors = copy.deepcopy(node.output_tensors)
                                pre_node.output_tensors[0].dest_op.append(model.nodes[-1].name)
                                for t in model.nodes[-1].input_tensors:
                                    if t.name == t_name:
                                        t.name = node.output_tensors[0].name
                                        break
                                remove_node_names.append(node.name)
            model.remove_nodes(remove_node_names)

        def _revert_logits_output_dtype(model, graph_dtype):
            for t in model.nodes[-1].input_tensors:
                pre_node = model.get_node_by_name(t.source_op[0])
                if pre_node and EXECUTOR_TYPE.get(pre_node.op_type, pre_node.op_type) in \
                   ['InnerProduct', 'Softmax', 'LogSoftmax']:
                    if pre_node.attr.get('output_dtype', "fp32") == graph_dtype:
                        pre_node.attr['output_dtype'] = 'fp32'

        if util.get_autocast_info()['cast_type'] != "bf16":
            return model
        graph_dtype = _quantized_dtype(model)
        # all in and out tensors support the dtype
        bf16_op = [
            'InnerProduct', 'Slice', 'Matmul', 'Reshape', 'BinaryOp', 'BinaryAdd', 'Reorder',
            'Concat', 'Softmax', 'LayerNorm', 'LogSoftmax', 'Convolution', 'Gather'
        ]
        s8_op = ['InnerProduct', 'Reshape', 'Shape', 'BinaryOp']
        checker = {'bf16': bf16_op, 's8': s8_op}

        non_quantized_patterns = [[(0, 'Range'), (1, ['Div', 'BinaryOp']), (2, 'Pow'),
                                   (3, ['Div', 'BinaryOp']), (4, 'Reshape'),
                                   (7, ['Matmul', 'BatchMatmul'])],
                                  [(), (5, 'Range'), (6, 'Reshape'), (7, ['Matmul',
                                                                          'BatchMatMul'])]]
        match_ret = util.search_pattern(non_quantized_patterns, model)
        for ret in match_ret:
            mat_node = model.get_node_by_name(ret[-2])
            dst_ops = _get_dst_ops(mat_node, model)
            dst_ops_type = [n.op_type for n in dst_ops]
            if dst_ops_type == ['Sin', 'Cos'] or dst_ops_type == ['CosSin', 'CosSin']:
                for n in dst_ops:
                    model = _insert_quant_node(n, model, graph_dtype)

        if graph_dtype in ['bf16']:
            for node in model.nodes:
                if EXECUTOR_TYPE.get(node.op_type, node.op_type) in ['Quantize']:
                    dst_ops = _get_dst_ops(node, model)
                    for op in dst_ops:
                        if op.attr.get('output_dtype', 'fp32') != graph_dtype and \
                            EXECUTOR_TYPE.get(op.op_type, op.op_type) in checker[graph_dtype]:
                            op.attr['output_dtype'] = graph_dtype
                            _check_dst_op(op, model, checker, graph_dtype)
            _scan_nodes_graph_dtype(model, checker, graph_dtype)
            _remove_redundant_quant_node(model, graph_dtype)
            _revert_logits_output_dtype(model, graph_dtype)
            for t in model.nodes[0].output_tensors:
                if t.location and len(t.location) == 2:
                    break
                else:
                    if t.dtype == "fp32":
                        t.dtype = "bf16"

        return model



"""prune utils."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import yaml

from ..config import PrunerV2, WeightPruningConfig
from neural_compressor.conf.dotdict import DotDict

from .logger import logger

def check_config(prune_config):
    """Check if the configuration dict is valid for running Pruning object.

    Args:
        prune_config: A config dict object that contains Pruning parameters and configurations.

    Returns:
        None if everything is correct.

    Raises:
        AssertionError.
    """
    assert prune_config['start_step'] >= 0, "start_step should be greater than 0"
    assert prune_config['end_step'] >= -1, "end_step should be greater than 0"
    assert prune_config['end_step'] >= prune_config['start_step'], \
        "end_step should be greater than start_step"
    assert prune_config['target_sparsity'] >= 0 and prune_config['target_sparsity'] < 1.0, \
        "begin_pruning_step should be in range [0,1)"
    assert prune_config['pruning_frequency'] > 0, "pruning_frequency should be greater than 0"
    assert prune_config['max_sparsity_ratio_per_op'] >= 0 and prune_config['max_sparsity_ratio_per_op'] < 1, \
        "pruning_frequency should be greater than 0"
    assert prune_config['pruning_scope'] == "global" or prune_config['pruning_scope'] == "local", \
        "only support 'global' and 'local' prune domain"
    try:
        prune_config['resume_from_pruned_checkpoint'] = bool(prune_config['resume_from_pruned_checkpoint'])
    except:  # pragma: no cover
        assert False, "resume_from_pruned_checkpoint should be bool value"
    if "x" in prune_config["pattern"]:
        pattern = prune_config["pattern"].split('_')[-1].split('x')
        if pattern[0] == "channel" or pattern[1] == "channel":
            pass
        else:
            try:
                N = int(pattern[0])
                M = int(pattern[1])
            except:  # pragma: no cover
                assert False, "N or M can't convert to int"
            assert N > 0, "N should be greater than 0"
            assert M > 0, "M should be greater than 0"
    if ":" in prune_config["pattern"]:
        pattern = prune_config["pattern"].split('_')[-1].split(':')
        try:
            N = int(pattern[0])
            M = int(pattern[1])
        except:  # pragma: no cover
            assert False, "N or M can't convert to int"
        assert N > 0, "N should be greater than 0"
        assert M > N, "M should be greater than N"
        max_ratio = float(N) / M
        assert prune_config['target_sparsity'] <= max_ratio, \
            "in N:M pattern, the max sparsity is N/M={}".format(max_ratio)
        prune_config['max_sparsity_ratio_per_op'] = min(max_ratio, prune_config['max_sparsity_ratio_per_op'])
    if prune_config['reg_coeff'] != None:
        prune_config['reg_coeff'] = float(prune_config['reg_coeff'])
        assert prune_config['reg_coeff'] >= 0, "only support positive reg_type"
    assert prune_config["min_sparsity_ratio_per_op"] >= 0 and prune_config["min_sparsity_ratio_per_op"] <= \
           prune_config['max_sparsity_ratio_per_op'], \
        "min_sparsity_ratio_per_op should in[0, max_sparsity_ratio_per_op]"


def reset_none_to_default(obj, key, default):
    """Set undefined configurations to default values.

    Args:
        obj: A dict{key: value}
        key: A string representing the key in obj.
        default: When the key is not in obj, add key by the default item in original obj.
    """
    if obj == None:  # pragma: no cover
        return None
    if isinstance(obj, dict):
        if (not key in obj.keys()) or obj[key] == None:
            return default
        else:
            return obj[key]
    else:  # pragma: no cover
        if not hasattr(obj, key) or getattr(obj, key) == None:
            return default
        else:
            return getattr(obj, key)

def update_params(info):
    """Update parameters."""
    if "parameters" in info.keys():
        params = info["parameters"]
        for key in params:
            info[key] = params[key]


def process_weight_config(global_config, local_configs, default_config):
    """Process pruning configurations.

    Args:
        global_config: A config dict object that contains pruning parameters and configurations.
        local_config: A config dict object that contains pruning parameters and configurations.
        default_config: A config dict object that contains pruning parameters and configurations.

    Returns:
        pruners_info: A config dict object that contains pruning parameters and configurations.
    """
    pruners_info = []
    default_all = global_config
    for key in default_config.keys():
        default_all[key] = reset_none_to_default(default_all, key, default_config[key])

    if len(local_configs) == 0:  # pragma: no cover

        update_params(default_all)
        check_config(default_all)
        pruner_info = DotDict(default_all)
        pruners_info.append(pruner_info)
    else:  ##TODO need update, in this mode, we ingore the global op names
        for pruner_info in local_configs:
            for key in default_config.keys():
                ##pruner_info[key] = reset_none_to_default(pruner_info, key, global_config[key])
                pruner_info[key] = reset_none_to_default(pruner_info, key, default_all[key])
            update_params(pruner_info)
            check_config(pruner_info)
            pruner_info = DotDict(pruner_info)
            pruners_info.append(pruner_info)

    return pruners_info


def process_yaml_config(global_config, local_configs, default_config):
    """Process the yaml configuration file.

    Args:
        global_config: A config dict object that contains pruning parameters and configurations.
        local_config: A config dict object that contains pruning parameters and configurations.
        default_config: A config dict object that contains pruning parameters and configurations.

    Returns:
        pruners_info: A config dict object that contains pruning parameters and configurations.
    """
    pruners_info = []
    default_all = global_config
    for key in default_config.keys():
        default_all[key] = reset_none_to_default(default_all, key, default_config[key])
    if len(local_configs) == 0:
        update_params(default_all)
        check_config(default_all)
        pruner_info = DotDict(default_all)
        pruners_info.append(pruner_info)

    else:  ##TODO need update, in this mode, we ingore the global op names
        for pruner in local_configs:
            for key in default_config.keys():
                pruner_info = pruner.pruner_config
                pruner_info[key] = reset_none_to_default(pruner_info, key, default_all[key])
            update_params(pruner_info)
            check_config(pruner_info)
            pruner_info = DotDict(pruner_info)
            pruners_info.append(pruner_info)

    return pruners_info

def check_key_validity(template_config, user_config):
    """Check the validity of keys.

    Args:
        template_config: A default config dict object that contains pruning parameters and configurations.
        user_config: A user config dict object that contains pruning parameters and configurations.
    """
    def check_key_validity_dict(template_config, usr_cfg_dict):
        """Check the validity of keys in the dict..

        Args:
            template_config: A default config dict object that contains pruning parameters and configurations.
            usr_cfg_dict: A user config dict object that contains pruning parameters and configurations.
        """
        for user_key, user_value in usr_cfg_dict.items():
            if user_key not in template_config.keys():
                logger.warning(f"{user_key} is not supported for config")

    def check_key_validity_prunerv2(template_config, usr_cfg_dict):
        """Check the validity of keys in the prunerv2.

        Args:
            template_config: A default config dict object that contains pruning parameters and configurations.
            usr_cfg_dict: A user config dict object that contains pruning parameters and configurations.
        """
        for user_key, user_value in usr_cfg_dict.pruner_config.items():
            if user_key not in template_config.keys():
                logger.warning(f"{user_key} is not supported for config")
    
    # multi pruners
    if isinstance(user_config, list):
        for obj in user_config:
            if isinstance(obj, dict):
                check_key_validity_dict(template_config, obj)
            elif isinstance(obj, PrunerV2):
                check_key_validity_prunerv2(template_config, obj)
                
    # single pruner, weightconfig or yaml
    elif isinstance(user_config, dict):
        check_key_validity_dict(template_config, user_config)
    elif isinstance(user_config, PrunerV2):
        check_key_validity_prunerv2(template_config, user_config)
    return

def process_and_check_config(val):
    """Process and check configurations.
    
    Args:  
        val: A dict that contains the layer-specific pruning configurations.
    """
    default_global_config = {'target_sparsity': 0.9, 'pruning_type': 'snip_momentum', 'pattern': '4x1', 'op_names': [],
                             'excluded_op_names': [],
                             'start_step': 0, 'end_step': 0, 'pruning_scope': 'global', 'pruning_frequency': 1,
                             'min_sparsity_ratio_per_op': 0.0, 'max_sparsity_ratio_per_op': 0.98,
                             'sparsity_decay_type': 'exp',
                             'pruning_op_types': ['Conv', 'Linear'],
                             }
    default_local_config = {'resume_from_pruned_checkpoint': False, 'reg_type': None,
                            'criterion_reduce_type': "mean", 'parameters': {"reg_coeff": 0.0}}

    params_default_config = {"reg_coeff": 0.0}

    default_config = {}
    default_config.update(default_global_config)
    default_config.update(default_local_config)
    default_config.update(params_default_config)
    if isinstance(val, WeightPruningConfig):
        global_configs = val.weight_compression
        pruning_configs = val.pruning_configs
        check_key_validity(default_config, pruning_configs)
        check_key_validity(default_config, global_configs)
        return process_weight_config(global_configs, pruning_configs, default_config)
    else:
        val = val["pruning"]["approach"]["weight_compression_v2"]
        global_configs = val
        pruning_configs = val["pruners"]
        check_key_validity(default_config, pruning_configs)
        check_key_validity(default_config, global_configs)
        return process_yaml_config(global_configs, pruning_configs, default_config)

def process_config(config):
    """Obtain a config dict object from the config file.

    Args:
        config: A string representing the path to the configuration file.

    Returns:
        A config dict object.
    """
    if isinstance(config, str):
        try:
            with open(config, 'r') as f:
                content = f.read()
                val = yaml.safe_load(content)
                ##schema.validate(val)
            return process_and_check_config(val)
        except FileNotFoundError as f:  # pragma: no cover
            logger.error("{}.".format(f))
            raise RuntimeError(
                "The yaml file is not exist. Please check the file name or path."
            )
        except Exception as e:  # pragma: no cover
            logger.error("{}.".format(e))
            raise RuntimeError(
                "The yaml file format is not correct. Please refer to document."
            )

    if isinstance(config, WeightPruningConfig):
        return process_and_check_config(config)
    else:  # pragma: no cover
        assert False, f"not supported type {config}"


def parse_to_prune(config, model):
    """Keep target pruned layers.
    
    Args:
        config: A string representing the path to the configuration file.
        model: The model to be pruned.
    """
    modules = {}
    if config["op_names"] == None or config["op_names"] == []:  # pragma: no cover
        config["op_names"] = [".*"]
    for raw in config["op_names"]:
        try:
            pattern = re.compile(raw)
        except:  # pragma: no cover
            assert False, f"regular expression match does not support {raw}"
        for name, module in filter(lambda t: pattern.search(t[0]), model.named_modules()):
            for layer_type in config["pruning_op_types"]:
                if layer_type in type(module).__name__ and hasattr(module, "weight"):
                    modules[name] = module
                    break
    ##remove not to prune layers
    """Drop non-pruned layers."""
    exclude_names = config["excluded_op_names"]
    patterns = [re.compile(s) for s in exclude_names]
    if len(patterns) <= 0:
        return modules
    new_modules = {}
    for name in modules.keys():
        if any([p.search(name) for p in patterns]):
            continue
        new_modules[name] = modules[name]
    return new_modules


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine torch extractor file."""

from ..graph_utils import LazyImport
from .. import logger
from ..graph.graph import Graph
from ..ops.op import OPERATORS
from ..onnx_utils import graph_node_names_details
from ..graph_utils import names_from_input, get_data_dtype, construct_node
from ..ops.tensor import Tensor
from ..torch_utils import get_node_name, op_maps
from .. import graph_utils as util

torch = LazyImport('torch')

def removeUnusedNode(graph, unused_nodes):
    remove_list = []
    for node in graph.nodes():
        if node.kind() in unused_nodes:
            in_val = node.inputsAt(0)
            out_val = node.outputsAt(0)
            out_val.replaceAllUsesWith(in_val)
            remove_list.append(node)

    for node in remove_list:
        node.destroy()

def fuse_padding_seq(graph):
    old_g = """
            graph(%input_ids.1, %attention_mask.1, %3, %4, %5, %6, %7, %8, %9, %10):
                %11 : int = aten::size(%input_ids.1, %9)
                %12 : int[] = prim::ListConstruct(%11, %8)
                %attention_mask0.1 : Tensor = aten::view(%attention_mask.1, %12)
                %14 : Tensor = aten::slice(%attention_mask0.1, %9, %9, %7, %6)
                %15 : Tensor = aten::unsqueeze(%14, %6)
                %16 : Tensor = aten::unsqueeze(%15, %5)
                %17 : Tensor = aten::slice(%16, %4, %9, %7, %6)
                %18 : Tensor = aten::rsub(%17, %3, %6)
                %19 : Tensor = aten::mul(%18, %10)
                return (%19)
            """
    new_g = """
            graph(%input_ids.1, %attention_mask.1, %3, %4, %5, %6, %7, %8, %9, %10):
                %19 = aten::padding_sequence(%attention_mask.1, %10)
                return (%19)
            """
    torch._C._jit_pass_custom_pattern_based_rewrite_graph(old_g, new_g, graph)

class TorchExtractor(object):
    """The TorchExtractor class.

    Decorate the node in model.graph_def, and the new node has the attributes like input_tensors
    and output_tensors, these tensors record the source/dest op name. All of these nodes
    (in a list) will compose a graph, which is Graph class, as the return object.

    Args:
        model: Torchscript Model

    Return:
        Graph: Graph class, the new graph object
    """
    @classmethod
    def __call__(self, model):
        """The __call__ function of the extractor."""
        graph, _ = torch._C._jit_pass_lower_graph(model.graph, model._c)
        torch._C._jit_pass_dce(graph)
        torch._C._jit_pass_remove_inplace_ops(graph)
        # torch._C._jit_pass_lower_all_tuples(graph)
        torch._C._jit_pass_constant_propagation(graph)
        removeUnusedNode(graph, ['aten::dropout', 'prim::NumToTensor', 'aten::to', 'aten::contiguous',
                                 'aten::alias', 'aten::Int', 'aten::ScalarImplicit'])
        logger.info('Start to extarct torch model ops...')
        fuse_padding_seq(graph)
        new_graph = Graph()
        new_graph.framework_modeling_config['framework'] = 'torch'
        graph_nodes_dict = {}
        util.quant_info_init()

        # insert graph.inputs
        model_input_tensors = []
        for idx, in_tensor in enumerate(graph.inputs()):
            dest_ops = []
            for val_user in in_tensor.uses():
                next_node = val_user.user
                dest_ops.append(get_node_name(next_node))
            dtype = 'fp32'
            if in_tensor.debugName() == 'position_ids':
                continue  # TODO: remove this
            input_tensor = Tensor(name=in_tensor.debugName(),
                source_op=[],
                dest_op=dest_ops,
                shape=[-1, -1],
                data=None,
                dtype=dtype
                )
            graph_nodes_dict[in_tensor.debugName()] = input_tensor
            model_input_tensors.append(input_tensor)
        for idx, out_tensor in enumerate(graph.outputs()):
            new_graph.output_tensors_name.append(out_tensor.debugName())

        # parse weights
        for node in graph.nodes():
            if node.kind() == 'prim::Constant' and node.outputsAt(0).type().kind() == 'TensorType':
                out_val = node.outputsAt(0)
                tensor_name = out_val.debugName()
                out_tensor = out_val.toIValue()
                dest_ops = []
                for val_user in out_val.uses():
                    next_node = val_user.user
                    dest_ops.append(get_node_name(next_node))
                if out_tensor.dtype == torch.qint8:
                    # extrace min max from tensor
                    fp32_data = out_tensor.dequantize()
                    if out_tensor.qscheme() == torch.per_channel_affine or \
                       out_tensor.qscheme() == torch.per_channel_symmetric:
                        # per_channel case
                        fp32_min = torch.min(fp32_data, 1).values.numpy()
                        fp32_max = torch.max(fp32_data, 1).values.numpy()
                    else:
                        fp32_min = fp32_data.min().numpy()
                        fp32_max = fp32_data.max().numpy()
                    dtype = 's8' + "_weight"
                    util.insert_quant_info(tensor_name, [fp32_min, fp32_max, dtype])

                    # ensure weight is sym quantized.
                    out_tensor = out_tensor.int_repr()

                elif out_tensor.dtype == torch.quint8:
                    logger.error("Tensor {} of uint8 is not supported.".format(tensor_name))
                    import sys; sys.exit(1)

                if out_tensor.dtype == torch.float64:
                    out_tensor = out_tensor.to(torch.float32)
                weight = out_tensor.detach().numpy()
                weight_tensor = Tensor(name=tensor_name,
                    source_op=[],
                    dest_op=dest_ops,
                    shape=list(out_tensor.shape),
                    data=weight,
                    dtype=get_data_dtype(weight)
                    )
                graph_nodes_dict[tensor_name] = weight_tensor
        

        for in_tensor in model_input_tensors:
            if in_tensor.name.split('.')[0] in ['attention_mask', 'position_ids', 'input_ids']:
                in_tensor.dtype = 'int32'  # TODO: refine this
        input_data_node = construct_node('input_data',
                                            'Input',
                                            output_tensors=model_input_tensors)
        #new_graph.inputs_dict = graph_nodes_dict
        input_tensors_name = [item for item in graph_nodes_dict]
        new_graph.input_tensors_name = input_tensors_name
        for node in graph.nodes():
            if node.kind() not in op_maps:
                logger.error("node kind {} is not mapped.".format(node.kind()))
                import sys; sys.exit(1)
            op_type = op_maps[node.kind()]
            if op_type not in OPERATORS.keys():
                    op_type = "OpAny"
            new_node = OPERATORS[op_type]()
            new_node.extract('torch', node, model, graph_nodes_dict)
            new_graph.insert_nodes(len(new_graph.nodes), [new_node])

        new_graph.insert_nodes(0, [input_data_node])
        return new_graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RoraryPosEmb Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
import numpy as np
import copy


@pattern_registry(pattern_type='RoraryPosEmb')
class RoraryPosEmb(Pattern):
    """The RoraryPosEmb pattern.

    Fuse the original sub-graph into the custom acceleration 'RoraryPosEmb' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'RoraryPosEmb': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (3, 'View'), (4, 'Unsqueeze')],
                               [(), (1, 'View'), (2, 'Repeat'), (3, 'View')]
                                ],
                        'out': [[(0, 'Reshape'), (1, 'Gather')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 2,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },
                {
                    'patterns': {
                        'in': [[(0, 'Add'), (1, 'Slice'), (2, 'Unsqueeze'), 
                                (3, 'Slice')]
                                ],
                        'out': [[(0, 'Reshape'), (1, 'Slice')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 3,
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]],
                        1: [[{
                            'input_data': [2]
                        },
                             {
                            'input_data': [0]
                        }], [[1,2], 3]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },

                {
                    'patterns': {
                        'in': [[(0, 'Slice'), (1, 'Slice'), (2, 'Slice'), (3, 'Slice'), 
                                (4, 'Neg'), (5, 'Stack'), (6, 'Flatten')]
                                ],
                        'out': [[(0, 'Slice'), (1, 'Mul'), (2, 'Reshape'), (3, 'Concat'), (4, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                        1: 4,
                        2: 0,
                        3: 5,
                        4: 6
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{
                            5: [1]
                        }], [[1], 2]],
                        4: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            6: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 3, 5]
                },
                
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (2, 'Add'), (3, 'Arange')],
                               [(), (1, 'Shape'), (2, 'Add')]
                                ],
                        'out': [[(0, 'Range')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3,
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        },{
                            'input_data': [2]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                }
            ]
        }
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                in_reshape_node = model.get_node_by_name(new_node_names[i][0])
                in_reshape_node.attr = OrderedDict({"dst_shape" : "-1, 1"})
                gather_node = model.get_node_by_name(new_node_names[i][1])
                gather_node.attr = OrderedDict({'axis': '0', 'batch_dims': '1'})
                idx_value = np.array([0, 0], dtype=np.int32)
                idx_tensor = Tensor(name=new_node_names[i][1] + "_idx",
                                        source_op=[],
                                        dest_op=[new_node_names[i][1]],
                                        data = idx_value,
                                        shape = [2],
                                        dtype="int32")
                gather_node.input_tensors.insert(0, idx_tensor)
            
            #  batch_dims: 0
                
        pattern_dict = pattern_mapping_config['RoraryPosEmb'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("RoraryPosEmb", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr(new_node_names, ret_old_nodes, model)
            

        def _set_attr1(new_node_names, ret_old_nodes, model):
            remove_shape_list = []
            for i in range(len(new_node_names)):
                for in_tensor in ret_old_nodes[i][0].input_tensors:
                    shape_node = model.get_node_by_name(in_tensor.source_op[0])
                    if len(shape_node.output_tensors[0].dest_op) == 0:
                        remove_shape_list.append(shape_node.name)
                reshape_node = model.get_node_by_name(new_node_names[i][0])
                reshape_node.attr = OrderedDict({"dst_shape" : "1, -1, 1, 64"})
                slice_node = model.get_node_by_name(new_node_names[i][1])
                slice_node.attr = OrderedDict({"starts_with_tensor" : "1",
                                               "ends_add" : "1", "axes" : "1", "steps" : "1"})
            model.remove_nodes(remove_shape_list)
                
        pattern_dict = pattern_mapping_config['RoraryPosEmb'][1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("RoraryPosEmb", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr1(new_node_names, ret_old_nodes, model)

        def _set_attr2(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                slice_node = model.get_node_by_name(new_node_names[i][0])
                slice_node.attr = ret_old_nodes[i][1].attr
                neg_value = np.array([-1], dtype=np.float32)
                neg_tensor = Tensor(name=new_node_names[i][1] + "_neg",
                                        source_op=[],
                                        dest_op=[new_node_names[i][1]],
                                        data = neg_value,
                                        shape = [1],
                                        dtype="fp32")
                mul_node = model.get_node_by_name(new_node_names[i][1])
                mul_node.input_tensors.append(neg_tensor)
                mul_node.attr = OrderedDict({'algorithm': 'mul'})
                old_slice_node = model.get_node_by_name(ret_old_nodes[i][0].input_tensors[0].source_op[0])
                concat_node = model.get_node_by_name(new_node_names[i][3])
                first_slice_node = model.get_node_by_name(concat_node.input_tensors[1].source_op[0])
                old_slice_node.output_tensors[0].dest_op.append(first_slice_node.name)
                first_slice_node.input_tensors[0] = old_slice_node.output_tensors[0]
                concat_node.attr = OrderedDict({'axis': '4'})
                reshape_node = model.get_node_by_name(new_node_names[i][2])
                reshape_node.attr = OrderedDict({'unsqueeze': '-1'})
                
                reshape_output = Tensor(name=concat_node.input_tensors[1].name + "_reshape",
                                        source_op=[concat_node.name + "_reshape"],
                                        dest_op=[concat_node.name],
                                        dtype=concat_node.input_tensors[1].dtype)
                
                reshape_op = util.construct_node(
                    node_name=concat_node.name + "_reshape",
                    op_type='Reshape',
                    input_tensors=[concat_node.input_tensors[1]],
                    output_tensors=[reshape_output],
                    attr=OrderedDict({'unsqueeze': '-1'}))
                concat_node.input_tensors[1].source_op = [reshape_op.name]
                concat_node.input_tensors[1] = reshape_output
                insert_idx = model.get_node_id(new_node_names[i][3])
                model.insert_nodes(insert_idx, [reshape_op])
                
                last_reshape_node = model.get_node_by_name(new_node_names[i][4])
                last_reshape_node.attr = OrderedDict({'mul': '3, 4'})
        pattern_dict = pattern_mapping_config['RoraryPosEmb'][2]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("RoraryPosEmb", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr2(new_node_names, ret_old_nodes, model)
            
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The InnerproductReshapeFusion Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='InnerproductReshapeFusion')
class InnerproductReshapeFusion(Pattern):
    """The InnerproductReshapeFusion pattern.

    Fuse the original sub-graph into the custom acceleration 'InnerproductReshapeFusion' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'InnerproductReshapeFusion': [
                {
                    'patterns': {
                        'in': [[(0, 'InnerProduct'), (1, 'Shape'), (3, 'View')],
                                [(0, 'InnerProduct'), (2, 'Shape'), (3, 'View')],
                                [(0, 'InnerProduct'), (3, 'View')]
                                ],
                        'out': [[(0, 'InnerProduct')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            'input_data': [0]
                        }
                        ], [[0, 1, 2], 3]]
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [3, 0]
                },
            ]
        }



       
        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                mat_node_idx = model.get_node_id(new_node_names[i][0])
                ret_mat_node = ret_old_nodes[i][1]
                if len(ret_mat_node.input_tensors) == 3:
                    model.nodes[mat_node_idx].input_tensors.insert(-1, copy.deepcopy(
                        ret_mat_node.input_tensors[-1]))
                attr = OrderedDict()
                if 'shape' in ret_old_nodes[i][0].attr.keys():
                    attr['reshape'] = ret_old_nodes[i][0].attr['shape']
                    attr['reshape_dims'] = '0'
                model.nodes[mat_node_idx].attr = attr
        
        if model.framework_modeling_config['framework'] == 'torch':
            pattern_dict = pattern_mapping_config['InnerproductReshapeFusion'][0]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("InnerproductReshapeFusion",
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                _set_attr(new_node_names, ret_old_nodes, model)

                return model

        return model



import numpy as np
import os
import shutil
import tensorflow as tf
import unittest
from datasets import load_dataset, load_metric
from intel_extension_for_transformers.optimization import (
    metrics,
    objectives,
    QuantizationConfig,
    TFOptimization
)
# from intel_extension_for_transformers.optimization import metrics, objectives
from transformers import (
    AutoTokenizer,
    DefaultDataCollator,
    HfArgumentParser,
    TFAutoModelForSequenceClassification,
    TFTrainingArguments,
)

os.environ["WANDB_DISABLED"] = "true"


class TestTFQuantization(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = TFAutoModelForSequenceClassification.from_pretrained(
            'bhadresh-savani/distilbert-base-uncased-sentiment-sst2'
        )
        raw_datasets = load_dataset("glue", "sst2")["validation"]
        tokenizer = AutoTokenizer.from_pretrained("bhadresh-savani/distilbert-base-uncased-sentiment-sst2")
        non_label_column_names = [name for name in raw_datasets.column_names if name != "label"]
        def preprocess_function(examples):
            # Tokenize the texts
            args = (
                (examples["sentence"],)
            )
            result = tokenizer(*args, padding=True, max_length=64, truncation=True)

            return result
        raw_datasets = raw_datasets.map(preprocess_function, batched=True, load_from_cache_file=False)
        data_collator = DefaultDataCollator(return_tensors="tf")
        dataset = raw_datasets.select(range(10))
        self.dummy_dataset = dataset.to_tf_dataset(
            columns=[col for col in dataset.column_names if col not in
                     set(non_label_column_names + ["label"])],
            shuffle=False,
            batch_size=2,
            collate_fn=data_collator,
            drop_remainder=False,
            # `label_cols` is needed for user-defined losses, such as in this example
            # datasets v2.3.x need "labels", not "label"
            label_cols=["labels"] if "label" in dataset.column_names else None,
        )


    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp', ignore_errors=True)
        shutil.rmtree('./quantized_model', ignore_errors=True)

    def test_tf_model_quant(self):
        parser = HfArgumentParser(TFTrainingArguments)
        args = parser.parse_args_into_dataclasses(args=["--output_dir", "./quantized_model",
                                                        "--per_device_eval_batch_size", "2"])
        metric = load_metric("glue", "sst2")
        def compute_metrics(preds, label_ids):
            preds = preds["logits"]
            preds = np.argmax(preds, axis=1)
            result = metric.compute(predictions=preds, references=label_ids)
            if len(result) > 1:
                result["combined_score"] = np.mean(list(result.values())).item()
            return result
        self.optimizer = TFOptimization(
            model=self.model,
            args=args[0],
            compute_metrics=compute_metrics
        )
        tune_metric = metrics.Metric(
            name="accuracy", greater_is_better=True, is_relative=False, criterion=0.5
        )
        quantization_config = QuantizationConfig(
            framework="tensorflow",
            approach="POSTTRAININGSTATIC",
            metrics=[tune_metric],
            objectives=[objectives.performance]
        )
        quantized_model = self.optimizer.quantize(quant_config=quantization_config,
            train_dataset=self.dummy_dataset, eval_dataset=self.dummy_dataset)
        loaded_model = tf.saved_model.load(args[0].output_dir)

        def eval_func(model):
            return 1

        def train_func(model):
            return model

        self.optimizer.quantize(quant_config=quantization_config,
                                train_func=train_func,
                                eval_func=eval_func)

        quantization_config = QuantizationConfig(
            framework="tensorflow",
            approach="POSTTRAININGSTATIC",
            metrics=[tune_metric],
            objectives=[objectives.performance],
            recipes={"first_conv_or_matmul_quantization": True,
                     "last_conv_or_matmul_quantization": True,
                     }
        )
        self.optimizer.quantize(quant_config=quantization_config,
                                train_func=train_func,
                                eval_func=eval_func)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import unittest

from datasets import load_dataset
from intel_extension_for_transformers.preprocessing.data_augmentation import DataAugmentation


def build_fake_dataset(save_path):
    from datasets import load_dataset
    from intel_extension_for_transformers.preprocessing.utils import EOS

    split = 'validation'
    count = 10
    dataset = load_dataset('glue', 'sst2', split='validation')
    origin_data = os.path.join(save_path, split + '.csv')
    print("original data:")
    with open(origin_data, 'w') as fw:
        fw.write('label' + '\t' + 'sentence' + '\n')
        for d in dataset:
            fw.write(str(d['label']) + '\t' + d['sentence'] + EOS + '\n')
            print(str(d['label']) + '\t' + d['sentence'] + EOS + '\n')
            count -= 1
            if count == 0:
                break
    return origin_data


class TestDataAugmentation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        os.makedirs("test_data", exist_ok=True)
        self.result_path = "test_data"
        self.origin_data = build_fake_dataset(self.result_path)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.result_path, ignore_errors=True)

    def test_text_generation_augmentation(self):
        aug = DataAugmentation(augmenter_type="TextGenerationAug")
        aug.input_dataset = self.origin_data
        aug.output_path = os.path.join(self.result_path, "test1.cvs")
        aug.augmenter_arguments = {'model_name_or_path': 'gpt2-medium'}
        aug.data_augment()
        print("Augmented data:")
        count = 0
        with open(aug.output_path, encoding='utf8') as f:
            for line in f:
                count += 1
                print(line)
        print("count:", count)
        self.assertTrue(count == 11)

    def test_keyboard_augmentation(self):
        aug = DataAugmentation(augmenter_type="KeyboardAug")
        aug.input_dataset = self.origin_data
        aug.column_names = "sentence"
        aug.output_path = os.path.join(self.result_path, "test2.cvs")
        aug.data_augment()
        raw_datasets = load_dataset("csv", data_files=aug.output_path, delimiter="\t", split="train")
        self.assertTrue(len(raw_datasets) == 10)

    def test_ocr_augmentation(self):
        aug = DataAugmentation(augmenter_type="OcrAug")
        aug.input_dataset = self.origin_data
        aug.column_names = "sentence"
        aug.output_path = os.path.join(self.result_path, "test2.cvs")
        aug.data_augment()
        raw_datasets = load_dataset("csv", data_files=aug.output_path, delimiter="\t", split="train")
        self.assertTrue(len(raw_datasets) == 10)

    def test_spelling_augmentation(self):
        aug = DataAugmentation(augmenter_type="SpellingAug")
        aug.input_dataset = self.origin_data
        aug.column_names = "sentence"
        aug.output_path = os.path.join(self.result_path, "test2.cvs")
        aug.data_augment()
        raw_datasets = load_dataset("csv", data_files=aug.output_path, delimiter="\t", split="train")
        self.assertTrue(len(raw_datasets) == 10)

    def test_contextualwordembsforsentence_augmentation(self):
        aug = DataAugmentation(augmenter_type="ContextualWordEmbsForSentenceAug")
        aug.input_dataset = self.origin_data
        aug.column_names = "sentence"
        aug.output_path = os.path.join(self.result_path, "test2.cvs")
        aug.augmenter_arguments = {"model_path": "xlnet-base-cased"}
        aug.data_augment()
        raw_datasets = load_dataset("csv", data_files=aug.output_path, delimiter="\t", split="train")
        self.assertTrue(len(raw_datasets) == 10)


if __name__ == "__main__":
    unittest.main()



"""Setup and install modules."""
from io import open
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
import os
import re
import sys
import shutil
from packaging.version import Version
from subprocess import check_output

cwd = os.path.dirname(os.path.abspath(__file__))

try:
    filepath = './intel_extension_for_transformers/version.py'
    with open(filepath) as version_file:
        __version__, = re.findall('__version__ = "(.*)"', version_file.read())
except Exception as error:
    assert False, "Error: Could not open '%s' due %s\n" % (filepath, error)

# define package data
package_data_dict = {
            '': ['*.py', '*.yaml'],
            'executor': ['*.py'],
        }

# define install requirements
install_requires_list = ['packaging', 'numpy', 'schema']
bac_install_requires_list = ['pyyaml', 'tqdm']
opt_install_requires_list = ['transformers>=4.12.0']

# define scripts
scripts_list = ['intel_extension_for_transformers/backends/neural_engine/bin/neural_engine']

# --develop: only install backends
only_install_backends = False
if "--backends" in sys.argv:
    only_install_backends = True
    sys.argv.remove("--backends")

if only_install_backends:
    project_name = "intel_extension_for_transformers_backends"
    packages_list = ["intel_extension_for_transformers", "intel_extension_for_transformers.backends"]
    bac_packages_list = find_packages("intel_extension_for_transformers/backends")
    bac_packages_list = ["intel_extension_for_transformers.backends." + i for i in bac_packages_list]
    packages_list.extend(bac_packages_list)
    install_requires_list.extend(bac_install_requires_list)
else:
    project_name = "intel_extension_for_transformers"
    packages_list = find_packages()
    install_requires_list.extend(opt_install_requires_list)

def which(thefile):
    """Get the path where the file is located."""
    path = os.environ.get("PATH", os.defpath).split(os.pathsep)
    if path == None:
        return None
    for d in path:
        fname = os.path.join(d, thefile)
        fnames = [fname]
        if sys.platform == 'win32':
            exts = os.environ.get('PATHEXT', '').split(os.pathsep)
            if exts == None:
                return None
            fnames += [fname + ext for ext in exts]
        for name in fnames:
            if os.access(name, os.F_OK | os.X_OK) and not os.path.isdir(name):
                return name
    return None


def get_version(cmd):
    """Returns cmake version."""
    try:
        for line in check_output([cmd, '--version']).decode('utf-8').split('\n'):
            if 'version' in line:
                print(line.strip().split(' ')[2])
                return Version(line.strip().split(' ')[2])
    except Exception as error:
        return Version('0.0.0')


def get_cmake_command():
    """Returns cmake command."""
    cmake_command = 'cmake'
    if sys.platform == 'win32':
        return cmake_command
    cmake3 = which('cmake3')
    cmake = which('cmake')
    if cmake3 is not None:
        if cmake is not None:
            bare_version = get_version('cmake')
            if (bare_version < Version("3.12.0") and get_version('cmake3') > bare_version):
                cmake_command = 'cmake3'
        else:
            cmake_command = 'cmake3'
    elif cmake is None:
        raise RuntimeError('no cmake or cmake3 found')
    return cmake_command


class build_ext(build_ext):
    """Extension builder."""
    def build_extension(self, ext):
        """Build the neural engine extension."""
        if not sys.platform.startswith("win"):
            import pathlib
            cwd = pathlib.Path().absolute()

            build_temp = pathlib.Path(self.build_temp)
            build_temp.mkdir(parents=True, exist_ok=True)
            extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
            executable_path = extdir.parent.absolute()
            executable_path.mkdir(parents=True,exist_ok=True)
            cmake_args = [
                '-DCMAKE_BUILD_TYPE=Release',
                '-DNE_WITH_SPARSELIB=ON',
                '-DNE_WITH_TESTS=OFF',
                '-DPYTHON_EXECUTABLE={}'.format(sys.executable)
            ]

            build_args = ['-j']
            cmake_command = get_cmake_command()
            os.chdir(str(build_temp))
            self.spawn([cmake_command, ext.sourcedir] + cmake_args)
            self.spawn(['make'] + build_args)

            import glob
            import shlex
            import subprocess
            bin_lists=glob.glob('bin/neural_engine*')
            bin_lists.extend(glob.glob('lib/*.so*'))
            for path in bin_lists:
                command = f'cp -d {path} {executable_path}'
                args = shlex.split(command)
                p=subprocess.Popen(args)
                p.wait()
                print(path)
            os.chdir(str(cwd))
        else:
            import pathlib
            cwd = pathlib.Path().absolute()
    
            build_temp = pathlib.Path(self.build_temp)
            build_temp.mkdir(parents=True, exist_ok=True)
            extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
            executable_path = extdir.parent.absolute()
            executable_path.mkdir(parents=True,exist_ok=True)
            cmake_args = [
                '-G Ninja',
                '-DCMAKE_BUILD_TYPE=Release',
                '-DNE_WITH_SPARSELIB=ON',
                '-DNE_WITH_TESTS=OFF',
                "-DCMAKE_C_COMPILER=cl.exe",
                "-DCMAKE_CXX_COMPILER=cl.exe",
                '-DEIGEN_BUILD_DOC=OFF',
                '-DPYTHON_EXECUTABLE={}'.format(sys.executable)
            ]
            build_args = [
                '--build',
                './',
                '-j'
            ]
            cmake_command = get_cmake_command() 
            os.chdir(str(build_temp))
            self.spawn([cmake_command, ext.sourcedir] + cmake_args)
            self.spawn(['cmake'] + build_args)
            import glob
            
            bin_lists=glob.glob('bin/*.exe')
            bin_lists.extend(glob.glob('bin/*.dll'))
            bin_lists.extend(glob.glob('lib/*.pyd'))
            for path in bin_lists:
                shutil.copy(path, executable_path, follow_symlinks=False)
            os.chdir(str(cwd))


class CMakeExtension(Extension):
    """CMakeExtension class."""
    def __init__(self, name, sourcedir=""):
        """Init a CMakeExtension object."""
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


def check_submodules():
    """Check submodules information."""
    def check_for_files(folder, files):
        if not any(os.path.exists(os.path.join(folder, f)) for f in files):
            report("Could not find any of {} in {}".format(", ".join(files), folder))
            report("Did you run 'git submodule update --init --recursive'?")
            sys.exit(1)

    def not_exists_or_empty(folder):
        return not os.path.exists(folder) or (os.path.isdir(folder)
                                              and len(os.listdir(folder)) == 0)

    git_modules_path = os.path.join(cwd, ".gitmodules")
    with open(git_modules_path) as f:
        folders = [
            os.path.join(cwd,
                         line.split("=", 1)[1].strip()) for line in f.readlines()
            if line.strip().startswith("path")
        ]

    # If none of the submodule folders exists, try to initialize them
    if all(not_exists_or_empty(folder)
           for folder in folders) and not sys.platform.startswith("win"):
        try:
            print(' --- Trying to initialize submodules')
            start = time.time()
            subprocess.check_call(["git", "submodule", "update", "--init", "--recursive"], cwd=cwd)
            end = time.time()
            print(' --- Submodule initialization took {:.2f} sec'.format(end - start))
        except Exception:
            print(' --- Submodule initalization failed')
            print('Please run:\n\tgit submodule update --init --recursive')
            sys.exit(1)


if __name__ == '__main__':
    check_submodules()

    setup(
        name = project_name,
        version = __version__,
        author = "Intel AIA/AIPC Team",
        author_email =
        "feng.tian@intel.com, haihao.shen@intel.com,hanwen.chang@intel.com, penghui.cheng@intel.com",
        description = "Repository of Intel® Intel Extension for Transformers",
        long_description = open("README.md", "r", encoding='utf-8').read(),
        long_description_content_type =" text/markdown",
        keywords =
        'quantization, auto-tuning, post-training static quantization, post-training dynamic quantization, quantization-aware training, tuning strategy',
        license = 'Apache 2.0',
        url="https://github.com/intel/",
        ext_modules = [CMakeExtension("neural_engine_py", str(cwd) + '/intel_extension_for_transformers/backends/neural_engine/')],
        packages = packages_list,
        include_package_data = True,
        package_dir = {'':'.'},
        package_data = package_data_dict,
        cmdclass = {
            'build_ext': build_ext,
        },
        install_requires = install_requires_list,
        scripts = scripts_list,
        python_requires = '>=3.7.0',
        classifiers = [
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'License :: OSI Approved :: Apache Software License',
        ],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for tensorflow framework."""

import os
import json
from collections import OrderedDict, UserDict
from neural_compressor.experimental import common

TMPPATH = os.path.join('tmp', 'model')
TEACHERPATH = os.path.join('tmp', 'teacher_model')
class TFDataloader(object):
    """Tensorflow dataloader.

    Args:
        dataset (string): Dataset
    """

    def __init__(self, dataset, batch_size=None):
        """Init an instance."""
        self.dataset = dataset
        self.batch_size = batch_size

    def __iter__(self):
        """Get the iteration of dataset."""
        for inputs, labels in self.dataset:
            if isinstance(inputs, dict) or isinstance(inputs, OrderedDict) \
                  or isinstance(inputs, UserDict):
                for name in inputs.keys():
                    inputs[name] = inputs[name].numpy()
            elif isinstance(inputs, list) or isinstance(inputs, tuple):
                inputs = [input.numpy() for input in inputs]
            else:
                inputs = inputs.numpy()

            if isinstance(labels, dict) or isinstance(labels, OrderedDict) \
                  or isinstance(labels, UserDict):   # pragma: no cover
                for name in labels.keys():
                    labels[name] = labels[name].numpy()
            elif isinstance(labels, list) or isinstance(labels, tuple):
                labels = [label.numpy() for label in labels]
            else:
                labels = labels.numpy()
            yield inputs, labels

    def __len__(self):
        """Return the length of dataset."""
        return len(self.dataset)


def distributed_init(worker_addresses, type='worker', index=0):
    """Init distribute environment.

    Args:
        worker_addresses: Addresses of all nodes.
        type: The type of node, such as worker.
        index: When index is 0, the node treat as a chief.
    """
    tf_config = {
        'cluster': {
            'worker': worker_addresses
        },
        'task': {'type': type, 'index': index}
    }
    os.environ['TF_CONFIG'] = json.dumps(tf_config)

def _is_chief(task_type, task_id):
    # here only consider the case in which TF_CONFIG task_type is set as worker
    # and task_id=0 represents the chief
    return (task_type == 'worker' and task_id == 0)

# get model folder path for the distributed environment
def get_filepath(base_dirpath, task_type, task_id):
    """Get model folder path for the distributed environment.

    Args:
        base_dirpath: The basic folder path.
        task_type: Task_type is set as worker.
        task_id: Task id. When task_id=0, the node treat as a chief.
    """
    if task_type is None:    # single node
        return base_dirpath
    elif _is_chief(task_type, task_id):
        return os.path.join(base_dirpath, 'chief')
    else:
        return os.path.join(base_dirpath, 'worker_' + str(task_id))


# convert a Keras model to SavedModel
def keras2SavedModel(model):   # pragma: no cover
    """Transfer keras model into save_model."""
    model = common.Model(model)
    return model.model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The TokenTypeEmbeddings pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='TokenTypeEmbeddings')
class TokenTypeEmbeddings(Pattern):
    """The TokenTypeEmbeddings pattern.

    Fuse the original sub-graph into the custom acceleration 'TokenTypeEmbeddings' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TokenTypeEmbeddings': [
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'Gather'), (4, 'Reshape')],
                               [(), (2, 'StridedSlice'), (3, 'Pack'), (4, 'Reshape')]],
                        'out': [[(0, 'Reshape'), (1, 'Gather'), (2, 'Reshape'), (3, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 'token_type_embeddings/after/reshape',
                        3: 4
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [1]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[1], 2]],
                        2: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [1]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1]
                },

                # bert_base_mrpc
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'OneHot'), (2, 'MatMul'), (6, 'Reshape')],
                               [(), (3, 'Shape'), (4, 'StridedSlice'), (5, 'Pack'),
                                (6, 'Reshape')]],
                        'out': [[(0, 'Reshape'), (1, 'OneHot'), (2, 'MatMul'), (3, 'Reshape'),
                                 (4, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 2,
                        3: 'token_type_embeddings/after/reshape',
                        4: 6
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [1]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            2: [1]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        4: [[{
                            'input_data': [1]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            6: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1, 2]
                },

                # geminet
                {
                    'patterns': {
                        'in': [[(0, 'Input'), (1, 'Gather'), (2, 'Add'), (3, 'LayerNorm')]],
                        'out': [[(0, 'Reshape'), (1, 'Gather'), (2, 'Reshape'), (3, 'Reshape'),
                                (4, 'Add'), (5, 'LayerNorm')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'token_type_embeddings/reshape',
                        1: 1,
                        2: 'token_type_embeddings/after/reshape',
                        3: 'token_type_embeddings/add_reshape',
                        4: 2,
                        5: 3,
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [1]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[1], 2]],
                        2: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        4: [[{
                            2:[0, 1]
                        }], [[0], 2]],
                        5: [[
                            {3: [1]}, {3: [2]}
                        ], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            2: [0]
                        }], [[0], 1]],
                        5: [[{
                            3: [0]
                        }], [[0], 1]],
                    },
                    'returns': [0, 1, 3]
                },

                # bert_base_sparse
                {
                    'patterns': {
                        'in': [[(0, 'Input'), (1, 'Gather'), (2, 'Add'), (3, 'Add')]],
                        'out': [[(0, 'Reshape'), (1, 'Gather'), (2, 'Reshape'),
                                (3, 'Reshape'), (4, 'Add'), (5, 'Add')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'token_type_embeddings/reshape',
                        1: 1,
                        2: 'token_type_embeddings/after/reshape',
                        3: 'token_type_embeddings/add_reshape',
                        4: 2,
                        5: 3,
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [1]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[1], 2]],
                        2: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [1]
                        }], [[1], 2]],
                        4: [[
                            {2: [0, 1]}],
                        [[0], 2]],
                        5: [[
                            {3: [0, 1]}],
                        [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            2: [0]
                        }], [[0], 1]],
                        5: [[{
                            3: [0]
                        }], [[0], 1]],
                    },
                    'returns': [1, 0]
                },
            ]
        }

        def _set_attr_with_gather(hidden_size, axis, batch_dims, node_names, model):
            attr1 = OrderedDict()
            attr1['dst_shape'] = -1
            attr2 = OrderedDict()
            attr2['axis'] = axis
            attr2['batch_dims'] = batch_dims
            attr3 = OrderedDict()
            attr3['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr3['dims'] = '0,1'
            attr4 = OrderedDict()
            attr4['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr4['dims'] = '0,1'
            attr4['mul'] = '1,2'

            reshape_0_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_0_node_idx].attr = attr1

            gather_node_idx = model.get_node_id(node_names[1])
            model.nodes[gather_node_idx].attr = attr2

            reshape_1_node_idx = model.get_node_id(node_names[2])
            model.nodes[reshape_1_node_idx].attr = attr3

            reshape_2_node_idx = model.get_node_id(node_names[3])
            model.nodes[reshape_2_node_idx].attr = attr4

        if model.framework_modeling_config['framework'] != 'onnxruntime':
            return model

        pattern_dict = pattern_mapping_config['TokenTypeEmbeddings'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("TokenTypeEmbeddings", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                gatherv2_node = ret_old_nodes[i][0]
                hidden_size = int(gatherv2_node.input_tensors[0].shape[-1])
                axis = gatherv2_node.attr['axis']
                batch_dims = gatherv2_node.attr['batch_dims']
                _set_attr_with_gather(hidden_size, axis, batch_dims, new_node_names[i], model)

            return model

        def _set_attr_with_onehot(hidden_size, one_hot_attr, mat_attr, node_names, model):
            attr1 = OrderedDict()
            attr1['dst_shape'] = -1
            attr2 = one_hot_attr
            attr3 = mat_attr
            attr4 = OrderedDict()
            attr4['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr4['dims'] = '0,1'
            attr5 = OrderedDict()
            attr5['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr5['dims'] = '0,1'
            attr5['mul'] = '1,2'
            for _ in range(len(node_names)):
                reshape_0_node_idx = model.get_node_id(node_names[0])
                model.nodes[reshape_0_node_idx].attr = attr1

                onehot_node_idx = model.get_node_id(node_names[1])
                model.nodes[onehot_node_idx].attr = attr2

                mat_node_idx = model.get_node_id(node_names[2])
                model.nodes[mat_node_idx].attr = attr3

                reshape_1_node_idx = model.get_node_id(node_names[3])
                model.nodes[reshape_1_node_idx].attr = attr4

                reshape_2_node_idx = model.get_node_id(node_names[4])
                model.nodes[reshape_2_node_idx].attr = attr5

        # bert_base onehot+matmul embeddings
        pattern_dict = pattern_mapping_config['TokenTypeEmbeddings'][1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("TokenTypeEmbeddings", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                onehot_node = ret_old_nodes[i][0]
                mat_node = ret_old_nodes[i][1]
                hidden_size = int(mat_node.input_tensors[1].shape[-1])
                transpose_a = mat_node.attr['transpose_a']
                transpose_b = mat_node.attr['transpose_b']
                mat_attr = OrderedDict()
                if transpose_a:
                    mat_attr['src0_perm'] = '1,0'
                # see OneDNN InnerProduct related requirements
                if not transpose_b:
                    mat_attr['src1_perm'] = '1,0'
                else:
                    mat_attr['src1_perm'] = '0,1'
                _set_attr_with_onehot(hidden_size, onehot_node.attr, mat_attr, new_node_names[i],
                                      model)

            return model

        # geminet
        pattern_dict = pattern_mapping_config['TokenTypeEmbeddings'][2]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("TokenTypeEmbeddings", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                gatherv2_node = ret_old_nodes[i][1]
                hidden_size = int(gatherv2_node.input_tensors[0].shape[-1])
                axis = gatherv2_node.attr['axis']
                batch_dims = gatherv2_node.attr['batch_dims']
                _set_attr_with_gather(hidden_size, axis, batch_dims, new_node_names[i], model)

                ln_node = ret_old_nodes[i][2]
                ln_1_node_idx = model.get_node_id(new_node_names[i][5])
                model.nodes[ln_1_node_idx].attr = ln_node.attr
                model.insert_nodes(0, [ret_old_nodes[i][0]])

            return model

        # bert_base_sparse
        pattern_dict = pattern_mapping_config['TokenTypeEmbeddings'][3]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("TokenTypeEmbeddings",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                gatherv2_node = ret_old_nodes[i][0]
                hidden_size = int(gatherv2_node.input_tensors[0].shape[-1])
                axis = gatherv2_node.attr['axis']
                batch_dims = gatherv2_node.attr['batch_dims']
                _set_attr_with_gather(hidden_size, axis, batch_dims, new_node_names[i], model)
                if len(ret_old_nodes[i]) == 2:
                    assert ret_old_nodes[i][1].op_type == 'Input'
                    model.insert_nodes(0, [ret_old_nodes[i][1]])
            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The InteractFeatures Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='InteractFeatures')
class InteractFeatures(Pattern):
    """The InteractFeatures pattern.

    Fuse the original sub-graph into the custom acceleration 'InteractFeatures' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'InteractFeatures': [
                # DLRM onnx model
                {
                    'patterns': {
                        'in': [[(0, 'Relu'), (1, 'Shape'), (2, 'Gather'), (3, 'Unsqueeze'), 
                                (7, 'Concat'), (9, 'Reshape')],
                               [(0, 'Relu'), (4, 'Shape'), (5, 'Gather'), (6, 'Unsqueeze'),
                                (7, 'Concat')],
                               [(0, 'Relu'), (8, 'Concat'), (9, 'Reshape')]],
                        'out': [[(0, 'Relu'), (1, 'Concat'), (2, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 8,
                        2: 9,
                    },
                    'input_tensors': {
                        0: [[{0: [0]}], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{1: [0]}], [[1], 2]], 
                    },
                    'output_tensors': {
                        0 : [[{0: [0]}], [[0], 1]],
                        1 : [[{8: [0]}], [[0], 1]],
                        2 : [[{9: [0]}], [[0], 1]],
                    },
                    'returns': [8, 9, 1, 4] # need modify the relu node source op
                },

                {
                    'patterns': {
                        'in': [[(0, 'MatMul'), (1, 'Shape'), (2, 'Gather'), (8, 'Concat'), \
                                (11, 'Reshape'), (13, 'Transpose'), (14, 'Reshape')],
                               [(1, 'Shape'), (3, 'Gather'), (4, 'Mul'), (5, 'Add'), \
                                (9, 'Gather'), (11, 'Reshape')],
                               [(0, 'MatMul'), (6, 'Transpose'), (7, 'Flatten'), (9, 'Gather')],
                               [(5, 'Add'), (10, 'Shape'), (12, 'Concat'), (14, 'Reshape')]
                        ],
                        'out': [[(0, 'Matmul'), (1, 'Reshape'), (2, 'Gather'), (3, 'Transpose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 7,
                        2: 9,
                        3: 14,
                    },
                    'input_tensors': {
                        0: [[{0: [0]}, {0: [1]}], [[0,1], 2]],
                        1: [[{0: [0]}], [[1], 2]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{14: [0]}], [[0], 1]],
                    },
                    'returns': [4, 5, 6, 9, 13]
                },

            ]
        }

        def _set_attr(concat_num, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = '-1, ' + str(concat_num) + ', -1'
            attr['dims'] = '0, 1'
            reshape_node_idx = model.get_node_id(node_names[2])
            model.nodes[reshape_node_idx].attr = attr


        concat_num = None
        for i in range(len(pattern_mapping_config['InteractFeatures'])):
            pattern_dict = pattern_mapping_config['InteractFeatures'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("InteractFeatures",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0 and i == 0:
                for j in range(len(new_node_names)):
                    concat_num = len(ret_old_nodes[j][0].input_tensors)
                    concat_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[concat_node_idx].input_tensors = ret_old_nodes[j][0].input_tensors
                    model.nodes[concat_node_idx].attr = ret_old_nodes[j][0].attr
                    _set_attr(concat_num, new_node_names[j], model)
                    relu_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[relu_node_idx].output_tensors[0].dest_op.remove(
                        ret_old_nodes[j][2].name)
                    model.nodes[relu_node_idx].output_tensors[0].dest_op.remove(
                        ret_old_nodes[j][3].name)
                
            elif len(new_node_names) != 0 and i == 1:
                for j in range(len(new_node_names)):
                    matmul_attr = OrderedDict()
                    matmul_attr['dst_perm'] = ret_old_nodes[j][2].attr['dst_perm']
                    matmul_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[matmul_node_idx].attr = matmul_attr
                    reshape_attr = OrderedDict()
                    reshape_attr['dst_shape'] = '-1, -1, -1'
                    reshape_attr['dims'] = '1, 1, 0'
                    reshape_attr['mul'] = '0, 1'
                    reshape_node_idx = model.get_node_id(new_node_names[j][1])
                    model.nodes[reshape_node_idx].attr = reshape_attr
                    gather_attr = OrderedDict()
                    gather_attr = ret_old_nodes[j][3].attr
                    gather_node_idx = model.get_node_id(new_node_names[j][2])
                    model.nodes[gather_node_idx].attr = gather_attr
                    gather_indices = concat_num * ret_old_nodes[j][0].input_tensors[0].data + \
                                     ret_old_nodes[j][1].input_tensors[0].data
                    ret_old_nodes[j][0].input_tensors[0].data = gather_indices
                    model.change_node_input_tensors(new_node_names[j][2], 0, \
                                        ret_old_nodes[j][0].input_tensors[0], mode='insert')
                    transpose_node_idx = model.get_node_id(new_node_names[j][3])
                    transpose_attr = OrderedDict()
                    transpose_attr = ret_old_nodes[j][4].attr
                    model.nodes[transpose_node_idx].attr = transpose_attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry, list2str
from .tensor import Tensor
import copy

@operator_registry(operator_type='Reorder')
class Reorder(Operator):
    """Parse the Transpose operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from onnxruntime."""
        if framework == "tensorflow":
            if self._input_tensors[1].source_op == []:
                dst_perm = list(self._input_tensors[1].data)
                self._attr['dst_perm'] = list2str(dst_perm)
                src_perm = [i for i in range(len(dst_perm))]
                self._attr['src_perm'] = list2str(src_perm)
                _ = self._input_tensors.pop()

        if framework == "onnxruntime":
            dst_perm = node.attribute[0].ints
            src_perm = [i for i in range(len(dst_perm))]
            self._attr['src_perm'] = list2str(src_perm)
            self._attr['dst_perm'] = list2str(dst_perm)

        if framework == 'torch':
            if node.kind() == 'aten::permute':
                # import pdb; pdb.set_trace()
                self._attr['src_perm'] = list2str(list(range(len(node.inputsAt(1).toIValue()))))
                self._attr['dst_perm'] = list2str(node.inputsAt(1).toIValue())
            if node.kind() == 'aten::transpose':
                dim0 = node.inputsAt(1).toIValue()
                dim1 = node.inputsAt(2).toIValue()
                self._attr['transpose_dims'] = list2str([dim0, dim1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np


@pattern_registry(pattern_type='MultiHeadAttention')
class MultiHeadAttention(Pattern):
    def __call__(self, model):
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        quant_info = util.get_quant_info()
        if not quant_info:
            return model
        pattern_mapping_config = {
            'MultiHeadAttention': [
                # for multi head attention
                # Bert based models
                {
                    'patterns': {
                        'in': [[(0, 'TransposeBatchMatMul'), (1, 'Softmax'),
                                (2, 'TransposeBatchMatMul')]],
                        'out': [[(0, 'MultiHeadAttention')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2,
                    },
                    'input_tensors': {
                        0: [[
                            {0: [0]},  # Q
                            {0: [1]},  # K
                            {2: [1]},  # V
                            {0: [2]},  # mask
                            {0: [3]},  # Q_min
                            {0: [4]},  # Q_max
                            {0: [5]},  # K_min
                            {0: [6]},  # K_max
                            {2: [4]},  # V_min
                            {2: [5]},  # V_max
                            {2: [2]},  # QK_min
                            {2: [3]},  # QK_max
                            {2: [6]},  # dst_min
                            {2: [7]},  # dst_max
                            ], 
                            [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13], 14]]
                    },
                    'output_tensors': {
                        0 : [[{2: [0]}], [[0], 1]]
                    },
                    'returns': [0, 1, 2]
                },

                # GPT-J torch model
                {
                    'patterns': {
                        'in': [[(0, ['Matmul', 'MatmulwithTranspose', 'BatchMatMul',
                                'TransposeBatchMatMul']), (1, ['Add', 'AddV2', 'BinaryAdd']),
                                (2, ['Div', 'BinaryOp']), (3, ['Add', 'AddV2', 'BinaryAdd']),
                                (4, 'Softmax'), (5, ['Matmul', 'MatmulwithTranspose',
                                'BatchMatMul', 'TransposeBatchMatMul'])]],
                        'out': [[(0, 'MultiHeadAttention')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 5,
                    },
                    'input_tensors': {
                        0: [[
                            {0: [0]},  # Q
                            {0: [1]},  # K
                            {5: [1]},  # V
                            {1: [1]},  # mask_0
                            {3: [1]},  # mask_1
                            {0: [2]},  # Q_min
                            {0: [3]},  # Q_max
                            {0: [4]},  # K_min
                            {0: [5]},  # K_max
                            {5: [4]},  # V_min
                            {5: [5]},  # V_max
                            {5: [2]},  # QK_min
                            {5: [3]},  # QK_max
                            {5: [6]},  # dst_min
                            {5: [7]},  # dst_max
                            ],
                            [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14], 15]]
                    },
                    'output_tensors': {
                        0 : [[{5: [0]}], [[0], 1]]
                    },
                    'returns': [0, 1, 2, 3, 4, 5]
                },
            ]
        }

        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                new_node = model.get_node_by_name(new_node_names[i][0])
                attr = OrderedDict()
                if len(ret_old_nodes[i]) == 3:
                    if 'src0_perm' in ret_old_nodes[i][0].attr.keys():
                        attr['Q_perm'] = ret_old_nodes[i][0].attr['src0_perm']
                    if 'src1_perm' in ret_old_nodes[i][0].attr.keys():
                        attr['K_perm'] = ret_old_nodes[i][0].attr['src1_perm']
                    if 'output_scale' in ret_old_nodes[i][0].attr.keys():
                        attr['output_scale'] = ret_old_nodes[i][0].attr['output_scale']
                    if 'src1_perm' in ret_old_nodes[i][2].attr.keys():
                        attr['V_perm'] = ret_old_nodes[i][2].attr['src1_perm']
                    if 'dst_perm' in ret_old_nodes[i][2].attr.keys():
                        attr['dst_perm'] = ret_old_nodes[i][2].attr['dst_perm']
                    if 'reshape' in ret_old_nodes[i][2].attr.keys():
                        attr['reshape'] = ret_old_nodes[i][2].attr['reshape']
                    if 'output_dtype' in ret_old_nodes[i][2].attr.keys():
                        attr['output_dtype'] = ret_old_nodes[i][2].attr['output_dtype']
                elif len(ret_old_nodes[i]) == 6:
                    if 'src0_perm' in ret_old_nodes[i][0].attr.keys():
                        attr['Q_perm'] = ret_old_nodes[i][0].attr['src0_perm']
                    if 'src1_perm' in ret_old_nodes[i][0].attr.keys():
                        attr['K_perm'] = ret_old_nodes[i][0].attr['src1_perm']
                    if ret_old_nodes[i][2].attr.get('algorithm', None) == 'div':
                        assert isinstance(ret_old_nodes[i][2].input_tensors[1].data, np.ndarray)
                        attr['output_scale'] = 1 / ret_old_nodes[i][2].input_tensors[1].data.item()
                    if 'src1_perm' in ret_old_nodes[i][5].attr.keys():
                        attr['V_perm'] = ret_old_nodes[i][5].attr['src1_perm']
                    if 'dst_perm' in ret_old_nodes[i][5].attr.keys():
                        attr['dst_perm'] = ret_old_nodes[i][5].attr['dst_perm']
                    if 'reshape' in ret_old_nodes[i][5].attr.keys():
                        attr['reshape'] = ret_old_nodes[i][5].attr['reshape']
                    if 'output_dtype' in ret_old_nodes[i][5].attr.keys():
                        attr['output_dtype'] = ret_old_nodes[i][5].attr['output_dtype']
                new_node.attr = attr
                if len(new_node.input_tensors) == 15:
                    mask_1 = new_node.input_tensors[4]
                    if model.get_node_by_name(mask_1.source_op[0]).op_type == "PaddingSequence":
                        new_node.input_tensors[3], new_node.input_tensors[4] = \
                            new_node.input_tensors[4], new_node.input_tensors[3]

        for i in range(len(pattern_mapping_config['MultiHeadAttention'])):
            pattern_dict = pattern_mapping_config['MultiHeadAttention'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("MultiHeadAttention",
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                _set_attr(new_node_names, ret_old_nodes, model)
                for node in model.nodes:
                    if node.op_type == 'PaddingSequence':
                        node.op_type = 'SequenceLength'
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Gelu Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='Gelu')
class Gelu(Pattern):
    """The Gelu pattern.

    Fuse the original sub-graph into the custom acceleration 'Gelu' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Gelu': [
                {
                    'patterns': {
                        'in': [[(0,'Pow'), (1, 'Mul'), (2, 'AddV2'), (3, 'Mul'), (4, 'Tanh'),
                                (5, 'AddV2'), (6, 'Mul'), (7, 'Mul')]],
                        'out': [[(0, 'Gelu')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 7
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            7: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },

                # distilbert_base
                {
                    'patterns': {
                        'in': [[(0, 'Div'), (1, 'Erf'), (2, 'Add'), (3, 'Mul'), (4, 'Mul')]],
                        'out': [[(0, 'Gelu')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 4
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': []
                },
            ]
        }

        for i in range(len(pattern_mapping_config['Gelu'])):
            pattern_dict = pattern_mapping_config['Gelu'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("Gelu", 
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    gelu_node = model.get_node_by_name(new_node_names[j][0])
                    attr = OrderedDict()
                    attr['algorithm'] = 'gelu_tanh'
                    gelu_node.attr = attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
from ..graph_utils import list2str

# Fused_op Reshape, ExpandDims+Sub+Mul
# This pattern is used for dealing with input_mask originally in bert model
@operator_registry(operator_type='PaddingSequence')
class PaddingSequence(Operator):
    """Register the PaddingSequence operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr."""
        if framework == "torch":
            self._attr['dst_shape'] = "-1,1,1,-1"
            self._attr['dims'] = 1
            self._attr['padding_value'] = node.inputsAt(1).toIValue().item()
            del self.input_tensors[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ArangewithReciprocal Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
import copy
import numpy as np

@pattern_registry(pattern_type='ArangewithReciprocal')
class ArangewithReciprocal(Pattern):
    """The ArangewithReciprocal pattern.

    Fuse the original sub-graph into the custom acceleration 'ArangewithReciprocal' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'ArangewithReciprocal': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Arange'), (2, 'Div'), (3, 'Pow'), 
                                (4, 'Reciprocal'), (5, 'Mul')]
                                ],
                        'out': [[(0, 'Range'), (1, 'Div'), (2, 'Pow'), 
                                (3, 'Div')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2,
                        2: 3,
                        3: 5
                        
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        }
                        ], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            3: [0]
                        }], [[0], 2]],
                        3: [[{
                            5: [1]
                        }
                        ], [[0], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },
                

                
            ]
        }


        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                
                slice_node = model.get_node_by_name(ret_old_nodes[i][0].input_tensors[0].source_op[0])
                fixed_pos_embedding_dim = int(slice_node.attr['ends'])
                range_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                attr['end'] = fixed_pos_embedding_dim
                attr['step'] = 2
                model.nodes[range_node_idx].attr = attr
                div_dim_node = model.get_node_by_name(new_node_names[i][1])
                div_dim = np.array([fixed_pos_embedding_dim], dtype=np.float32)
                div_dim_tensor = Tensor(name=div_dim_node.input_tensors[0].name + "_dim",
                                        source_op=[],
                                        dest_op=[div_dim_node.name],
                                        data = div_dim,
                                        shape = [1, 1],
                                        dtype="fp32")
                div_dim_node.input_tensors.append(div_dim_tensor)
                div_dim_node.attr = OrderedDict({'algorithm': 'div'})
                model.get_node_by_name(new_node_names[i][3]).attr = \
                                     OrderedDict({'algorithm': 'div'})
                reciprocal_node = model.get_node_by_name(new_node_names[i][3])
                reciprocal_node.input_tensors[0].data = np.array([1], dtype=np.float32)
                
        if model.framework_modeling_config['framework'] == 'torch':
            pattern_dict = pattern_mapping_config['ArangewithReciprocal'][0]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("ArangewithReciprocal", 
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                _set_attr(new_node_names, ret_old_nodes, model)
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
import numpy as np
from ..graph_utils import list2str


# tf.reshape(tensor, shape, name=None)
# param shape is s Tensor. Must be one of the following types: int32, int64.
# defines the shape of the output tensor.
@operator_registry(operator_type='PackagePositionEmbedding')
class PackagePositionEmbedding(Operator):
    """Parse the Reshape operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        #if framework == 'tensorflow':
        if self.input_tensors[1].source_op == []:
            self._attr['dst_shape'] = '1,-1,1024'
            self._attr['dims'] = '1'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RemoveConstantOP Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='RemoveConstantOP')
class RemoveConstantOP(Pattern):
    """The RemoveConstantOP pattern.

    Remove Constant OP
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        remove_list = []
        for node in model.nodes:
            if node.op_type == 'Constant':
                if node.output_tensors[0].name in model.input_tensors_name:
                    remove_list.append(node.name)
                    continue
                for dest_op_name in node.output_tensors[0].dest_op:
                    dest_node = model.get_node_by_name(dest_op_name)
                    new_input_tensors = []
                    for tensor in dest_node.input_tensors:
                        if tensor.name == node.output_tensors[0].name:
                            continue
                        new_input_tensors.append(copy.deepcopy(tensor))
                    dest_node.input_tensors = new_input_tensors
                remove_list.append(node.name)
        model.remove_nodes(remove_list)
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The LowerAllTuples Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='LowerAllTuples')
class LowerAllTuples(Pattern):
    """The LowerAllTuples pattern.

    LowerAllTuples
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        remove_list = []
        for node in model.nodes:
            if node.op_type in ['ListConstruct', 'TupleConstruct']:
                if node.output_tensors[0].dest_op == [] and node.output_tensors[0].name in model.output_tensors_name:
                    idx = model.output_tensors_name.index(node.output_tensors[0].name)
                    del model.output_tensors_name[idx]
                    for tensor in node.input_tensors:
                        for tensor in node.input_tensors:
                            model.output_tensors_name.insert(idx, tensor.name)
                            idx += 1

                for dest_op_name in node.output_tensors[0].dest_op:
                    dest_node = model.get_node_by_name(dest_op_name)
                    for i in range(len(dest_node.input_tensors)):
                        if dest_node.input_tensors[i].name == node.output_tensors[0].name:
                            del dest_node.input_tensors[i]
                            idx = i 
                            for tensor in node.input_tensors:
                                if node.name in tensor.dest_op:
                                    tensor.dest_op.remove(node.name)
                                tensor.dest_op.append(dest_node.name)
                                dest_node.input_tensors.insert(idx, copy.deepcopy(tensor))
                                idx += 1
                remove_list.append(node.name)
        node_idx = len(model.nodes) - 1
        while node_idx >= 0:
            node = model.nodes[node_idx]
            if node.op_type in ['ListUnpack', 'TupleUnpack']:
                for source_op_name in node.input_tensors[0].source_op:
                    source_op = model.get_node_by_name(source_op_name)
                    for i in range(len(source_op.output_tensors)):
                        if source_op.output_tensors[i].name == node.input_tensors[0].name:
                            del source_op.output_tensors[i]
                            idx = i
                            for tensor in node.output_tensors:
                                if node.name in tensor.source_op:
                                    tensor.source_op.remove(node.name)
                                tensor.source_op.append(source_op.name)
                                if source_op.op_type == 'Input':
                                    tensor.source_op = []
                                    tensor.dtype = "fp32"
                                    tensor.shape = [-1, -1, -1, -1]
                                source_op.output_tensors.insert(idx, copy.deepcopy(tensor))
                                idx += 1
                remove_list.append(node.name)
            node_idx -= 1

        model.remove_nodes(remove_list)
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The LastLayerShape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='LastLayerShape')
class LastLayerShape(Pattern):
    """The LastLayerShape pattern.

    Fuse the original sub-graph into the custom acceleration 'LastLayerShape' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'LastLayerShape': [
                {
                    'patterns': {
                        'in': [[(0, 'Pack'), (1, 'Reshape'), (2, 'Shape'), (3, 'StridedSlice'),
                                (5, 'Mul'), (6, 'Pack'), (7, 'Reshape'), (8, 'MatMulWithBias'),
                                (10, 'Reshape')], [(2, 'Shape'), (4, 'StridedSlice'), (5, 'Mul')],
                               [(3, 'StridedSlice'), (9, 'Pack'), (10, 'Reshape')]],
                        'out': [[(0, 'MatMulWithBias'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 8,
                        1: 10
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }, {
                            8: [1]
                        }, {
                            8: [2]
                        }], [[0, 1, 2], 3]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            10: [0]
                        }], [[0], 1]]
                    },
                    'returns': [8, 9]
                },
                {
                    'patterns': {
                        'in': [[(0, 'Pack'), (1, 'Reshape'), (2, 'StridedSlice'), (3, 'Squeeze')]],
                        'out': [[(0, 'Reshape'), (1, 'StridedSlice'), (2, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2,
                        2: 3
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 2]
                },

                # roberta_base / distil_bert_base
                {
                    'patterns': {
                        'in': [[(0, 'LayerNorm'), (1, 'Gather'), (2, 'MatMulWithBias')]],
                        'out': [[(0, 'LayerNorm'), (1, 'Reshape'), (2, 'StridedSlice'),
                                (3, 'Reshape'), (4, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 'last_layer_reshape',
                        2: 'last_layer_strided_slice',
                        3: 1,
                        4: 2
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]}, {0: [1]}, {0: [2]}
                        ], [[0, 1, 2], 3]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[
                            {2: [1]}, {2: [2]}
                        ], [[1, 2], 3]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{1: [0]}], [[0], 1]],
                        4: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 2]
                }
            ]
        }

        # bert_mlperf
        pattern_dict = pattern_mapping_config['LastLayerShape'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("LastLayerShape", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                mat_node = ret_old_nodes[i][0]
                pack_node = ret_old_nodes[i][1]
                last_dim = int(pack_node.input_tensors[-1].data)

                attr1 = mat_node.attr
                mat_node_idx = model.get_node_id(new_node_names[i][0])
                model.nodes[mat_node_idx].attr = attr1

                attr2 = OrderedDict()
                attr2['dst_shape'] = '-1,-1,' + str(last_dim)
                attr2['dims'] = '0,1'
                reshape_node_idx = model.get_node_id(new_node_names[i][1])
                model.nodes[reshape_node_idx].attr = attr2

            return model

        # bert_base_mrpc
        pattern_dict = pattern_mapping_config['LastLayerShape'][1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("LastLayerShape", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                pack_node = ret_old_nodes[i][0]
                strided_slice_node = ret_old_nodes[i][1]
                hidden_size = int(pack_node.input_tensors[-1].data)

                attr1 = OrderedDict()
                attr1['dst_shape'] = '-1,-1,' + str(hidden_size)
                attr1['dims'] = '0,1'
                reshape_0_node_idx = model.get_node_id(new_node_names[i][0])
                model.nodes[reshape_0_node_idx].attr = attr1

                attr2 = strided_slice_node.attr
                strided_slice_node_idx = model.get_node_id(new_node_names[i][1])
                model.nodes[strided_slice_node_idx].attr = attr2

                attr3 = OrderedDict()
                attr3['dst_shape'] = '-1,' + str(hidden_size)
                reshape_1_node_idx = model.get_node_id(new_node_names[i][2])
                model.nodes[reshape_1_node_idx].attr = attr3

            return model

        # roberta_base / distil_bert
        pattern_dict = pattern_mapping_config['LastLayerShape'][2]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("LastLayerShape", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                ln_node = ret_old_nodes[i][0]
                mat_node = ret_old_nodes[i][1]
                hidden_size = int(ln_node.input_tensors[-1].shape[-1])

                ln_node_idx = model.get_node_id(new_node_names[i][0])
                model.nodes[ln_node_idx].attr = ln_node.attr
                reshape_0_node_idx = model.get_node_id(new_node_names[i][1])
                model.nodes[reshape_0_node_idx].attr = OrderedDict({
                    'dst_shape': '-1,-1,' + str(hidden_size), 'dims': '0'})
                strided_slice_node_idx = model.get_node_id(new_node_names[i][2])
                model.nodes[strided_slice_node_idx].attr = OrderedDict({
                    'begin_mask': 5, 'ellipsis_mask': 0, 'end_mask': 5, 'new_axis_mask': 0,
                    'shrink_axis_mask': 0, 'begin': '0,0,0',  'end': '0,1,0', 'strides': '1,1,1'})
                reshape_1_node_idx = model.get_node_id(new_node_names[i][3])
                model.nodes[reshape_1_node_idx].attr = OrderedDict({
                        'dst_shape': '-1,' + str(hidden_size)})
                mat_node_idx = model.get_node_id(new_node_names[i][4])
                model.nodes[mat_node_idx].attr = mat_node.attr

            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Transformer2Dmodel_InstancenormReshapeFromConcat Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='Transformer2Dmodel_InstancenormReshapeFromConcat')
class Transformer2Dmodel_InstancenormReshapeFromConcat(Pattern):
    """The Transformer2Dmodel_InstancenormReshapeFromConcat pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_InstancenormReshapeFromConcat' graph.
    This pattern is used for the reshape node after the QKV.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Transformer2Dmodel_InstancenormReshapeFromConcat': [
                {
                    'patterns': {
                        'in': [[(0, 'Concat'), (1, 'Reshape'), (2, 'InstanceNormalization'),(4, 'Reshape')],
                        [(0, 'Concat'), (3, 'Shape'), (4, 'Reshape')]],
                        'out': [[(0, 'Concat'), (1, 'Reshape'),(2, 'InstanceNormalization'), (3, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 2,
                        3: 4
                    },
                    'input_tensors': {
                        0: [[{0: [0]}, {0: [1]}], [[0, 1], 2]],
                        1: [[{'input_data': [0]}], [[1], 2]],
                        2: [[{2: [1]}, {2: [2]}], [[1, 2], 3]],
                        3: [[{'input_data': [0]}], [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[{0: [0]}], [[0], 1]],
                        1: [[{1: [0]}], [[0], 1]],
                        2: [[{2: [0]}], [[0], 1]],
                        3: [[{4: [0]}], [[0], 1]],
                    },
                    'returns': [0, 1, 2, 3, 4]
                },
            ]
        }

        def _set_attr(batch_size, out_channel, height, weight, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = str(batch_size) + ',' + str(out_channel) + ',' + str(height) + ',' + str(weight)
            attr['dims'] = '0, 2, 3'

            reshape_node_idx = model.get_node_id(node_names[3])
            assert  model.nodes[reshape_node_idx].op_type == 'Reshape'
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['Transformer2Dmodel_InstancenormReshapeFromConcat'])):
            pattern_dict = pattern_mapping_config['Transformer2Dmodel_InstancenormReshapeFromConcat'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                "Transformer2Dmodel_InstancenormReshapeFromConcat", pattern_dict, model)

            print('i = ', i, 'Transformer2Dmodel_InstancenormReshapeFromConcat = ', new_node_names)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    continue
                    # first_ret_old_node = ret_old_nodes[j][0]
                    # assert first_ret_old_node.op_type == 'Div'
                    # div_node_idx = model.get_node_id(new_node_names[j][0])
                    # model.nodes[div_node_idx].attr = first_ret_old_node.attr

                    # second_ret_old_node = ret_old_nodes[j][1]
                    # assert second_ret_old_node.op_type == 'Reshape'
                    # new_node_idx = model.get_node_id(new_node_names[j][1])


                    # if second_ret_old_node.attr['dst_shape'].split(',')[0] == '0':
                    #     attr = OrderedDict()
                    #     shape_0 = '-1'
                    #     shape_1 = second_ret_old_node.attr['dst_shape'].split(',')[1]
                    #     shape_2 = second_ret_old_node.attr['dst_shape'].split(',')[2]
                    #     attr['dst_shape'] = shape_0 + ',' + shape_1 + ',' + shape_2
                    #     attr['dims'] = 0
                    #     model.nodes[new_node_idx].attr = attr
                    # else:
                    #     model.nodes[new_node_idx].attr = second_ret_old_node.attr

                    # # the third new node
                    # instanceNormalization_node_idx = model.get_node_id(new_node_names[j][2])
                    # model.nodes[instanceNormalization_node_idx].op_type = 'LayerNorm'
                    # attr_3 = OrderedDict()
                    # attr_3['epsilon'] = 9.999999960041972e-13
                    # model.nodes[instanceNormalization_node_idx].attr = attr_3

                    # forth_ret_old_node = ret_old_nodes[j][4]
                    # assert forth_ret_old_node.op_type == 'Reshape'
                    # #import pdb;pdb.set_trace()
                    # out_channel = 512
                    # #out_channel = first_ret_old_node.input_tensors[1].data.shape[0]
                    # new_node_idx = model.get_node_id(new_node_names[j][3])
                    # batch = model.nodes[new_node_idx].input_tensors[1].shape[0]
                    # height = model.nodes[new_node_idx].input_tensors[1].shape[2]
                    # weight = model.nodes[new_node_idx].input_tensors[1].shape[3]

                    # _set_attr(batch, out_channel, height, weight, new_node_names[j], model)
                return model

        return model
